import pytest
from aikido_zen.vulnerabilities.nosql_injection import detect_nosql_injection
from aikido_zen.context import Context
from collections import defaultdict, OrderedDict, ChainMap


@pytest.fixture
def create_context():
    def _create_context(
        query=None, headers=None, body=None, cookies=None, route_params=None
    ):
        class RequestContext:
            remote_address = "::1"
            method = "GET"
            url = "http://localhost:4000"
            query = {}
            headers = {}
            body = None
            cookies = {}
            route_params = {}
            source = "express"
            route = "/posts/:id"

        RequestContext.query = query if query else {}
        RequestContext.headers = headers if headers else {}
        RequestContext.body = body
        RequestContext.cookies = cookies if cookies else {}
        RequestContext.route_params = route_params if route_params else {}
        return RequestContext()

    return _create_context


def test_empty_filter_and_request(create_context):
    assert detect_nosql_injection(create_context(), {}) == {}


def test_ignore_if_filter_not_object(create_context):
    assert detect_nosql_injection(create_context(), "abc") == {}


def test_ignore_if_and_not_array(create_context):
    assert detect_nosql_injection(create_context(), {"$and": "abc"}) == {}


def test_ignore_if_or_not_array(create_context):
    assert detect_nosql_injection(create_context(), {"$or": "abc"}) == {}


def test_ignore_if_nor_not_array(create_context):
    assert detect_nosql_injection(create_context(), {"$nor": "abc"}) == {}


def test_ignore_if_nor_empty_array(create_context):
    assert detect_nosql_injection(create_context(), {"$nor": []}) == {}


def test_ignore_if_not_not_object(create_context):
    assert detect_nosql_injection(create_context(), {"$not": "abc"}) == {}


def test_filter_with_string_value_and_empty_request(create_context):
    assert detect_nosql_injection(create_context(), {"title": {"title": "title"}}) == {}


def test_filter_with_ne_and_empty_request(create_context):
    assert detect_nosql_injection(create_context(), {"title": {"$ne": None}}) == {}


def test_filter_with_string_value(create_context):
    assert (
        detect_nosql_injection(
            create_context(body={"title": {"title": "title", "a": "b"}}),
            {"title": {"title": "title", "a": "b"}},
        )
        == {}
    )


def test_using_gt_in_query_parameter(create_context):
    assert detect_nosql_injection(
        create_context(query={"title": {"$gt": ""}}), {"title": {"$gt": ""}}
    ) == {
        "injection": True,
        "source": "query",
        "pathToPayload": ".title",
        "payload": {"$gt": ""},
    }


def test_safe_filter(create_context):
    assert (
        detect_nosql_injection(
            create_context(query={"title": "title"}), {"$and": [{"title": "title"}]}
        )
        == {}
    )


def test_using_ne_in_body(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}), {"title": {"$ne": None}}
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_in_body_with_chainmap(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        ChainMap({"title": {"$ne": None}}, {}),
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_in_body_with_ordereddict(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        OrderedDict({"title": {"$ne": None}}),
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_in_body_different_name(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}), {"myTitle": {"$ne": None}}
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_in_headers_with_different_name(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}), {"someField": {"$ne": None}}
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_inside_and(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        {"$and": [{"title": {"$ne": None}}, {"published": True}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_inside_or(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        {"$or": [{"title": {"$ne": None}}, {"published": True}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_inside_nor(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        {"$nor": [{"title": {"$ne": None}}, {"published": True}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_inside_not(create_context):
    assert detect_nosql_injection(
        create_context(body={"title": {"$ne": None}}),
        {"$not": {"title": {"$ne": None}}},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".title",
        "payload": {"$ne": None},
    }


def test_using_ne_nested_in_body(create_context):
    assert detect_nosql_injection(
        create_context(body={"nested": {"nested": {"$ne": None}}}),
        {"$not": {"title": {"$ne": None}}},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".nested.nested",
        "payload": {"$ne": None},
    }


def test_using_ne_in_jwt_in_headers(create_context):
    assert detect_nosql_injection(
        create_context(
            # JWT token with the following payload:
            # {
            #   "sub": "1234567890",
            #   "username": {
            #     "$ne": null
            #   },
            #   "iat": 1516239022
            # }
            headers={
                "Authorization": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbWUiOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
            }
        ),
        {"username": {"$ne": None}},
    ) == {
        "injection": True,
        "source": "headers",
        "pathToPayload": ".Authorization<jwt>.username",
        "payload": {"$ne": None},
    }


def test_using_ne_in_jwt_in_bearer_header(create_context):
    assert detect_nosql_injection(
        create_context(
            # JWT token with the following payload:
            # {
            #   "sub": "1234567890",
            #   "username": {
            #     "$ne": null
            #   },
            #   "iat": 1516239022
            # }
            headers={
                "Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbWUiOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
            }
        ),
        {"username": {"$ne": None}},
    ) == {
        "injection": True,
        "source": "headers",
        "pathToPayload": ".Authorization<jwt>.username",
        "payload": {"$ne": None},
    }


def test_using_ne_in_jwt_in_cookies(create_context):
    assert detect_nosql_injection(
        create_context(
            # JWT token with the following payload:
            # {
            #   "sub": "1234567890",
            #   "username": {
            #     "$ne": null
            #   },
            #   "iat": 1516239022
            # }
            cookies={
                "session": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbWUiOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
            }
        ),
        {"username": {"$ne": None}},
    ) == {
        "injection": True,
        "source": "cookies",
        "pathToPayload": ".session<jwt>.username",
        "payload": {"$ne": None},
    }


def test_jwt_lookalike(create_context):
    assert (
        detect_nosql_injection(
            create_context(
                cookies={
                    "session": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbW!iOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
                }
            ),
            {"username": {"$ne": None}},
        )
        == {}
    )


def test_using_gt_in_query_parameter(create_context):
    assert detect_nosql_injection(
        create_context(query={"age": {"$gt": "21"}}), {"age": {"$gt": "21"}}
    ) == {
        "injection": True,
        "source": "query",
        "pathToPayload": ".age",
        "payload": {"$gt": "21"},
    }


def test_using_gt_in_query_parameter_with_other_params(create_context):
    assert detect_nosql_injection(
        create_context(query={"age": {"$gt": "21"}}),
        {"age": {"$gt": "21", "test": "true"}},
    ) == {
        "injection": True,
        "source": "query",
        "pathToPayload": ".age",
        "payload": {"$gt": "21"},
    }


def test_using_filter_as_body(create_context):
    assert detect_nosql_injection(
        create_context(body={"$gt": "21", "pswd": "Test"}),
        {"$gt": "21", "pswd": "Test"},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".",
        "payload": {"$gt": "21"},
    }


def test_using_gt_and_lt_in_query_parameter(create_context):
    assert detect_nosql_injection(
        create_context(body={"age": {"$gt": "21", "$lt": "100"}}),
        {"age": {"$gt": "21", "$lt": "100"}},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".age",
        "payload": {"$gt": "21", "$lt": "100"},
    }


def test_using_gt_and_lt_in_query_parameter_different_name(create_context):
    assert detect_nosql_injection(
        create_context(body={"age": {"$gt": "21", "$lt": "100"}}),
        {"myAge": {"$gt": "21", "$lt": "100"}},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".age",
        "payload": {"$gt": "21", "$lt": "100"},
    }


def test_using_gt_and_lt_in_query_parameter_nested(create_context):
    assert detect_nosql_injection(
        create_context(
            body={"nested": {"nested": {"age": {"$gt": "21", "$lt": "100"}}}}
        ),
        {"$and": [{"someAgeField": {"$gt": "21", "$lt": "100"}}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".nested.nested.age",
        "payload": {"$gt": "21", "$lt": "100"},
    }


def test_using_gt_and_lt_in_query_parameter_root(create_context):
    assert detect_nosql_injection(
        create_context(body={"$and": [{"someAgeField": {"$gt": "21", "$lt": "100"}}]}),
        {"$and": [{"someAgeField": {"$gt": "21", "$lt": "100"}}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".",
        "payload": {"$and": [{"someAgeField": {"$gt": "21", "$lt": "100"}}]},
    }


def test_where(create_context):
    assert detect_nosql_injection(
        create_context(body={"$and": [{"$where": "sleep(1000)"}]}),
        {"$and": [{"$where": "sleep(1000)"}]},
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".",
        "payload": {"$and": [{"$where": "sleep(1000)"}]},
    }


def test_array_body(create_context):
    assert detect_nosql_injection(
        create_context(
            body=[
                {
                    "$where": "sleep(1000)",
                },
            ]
        ),
        {
            "$and": [
                {
                    "$where": "sleep(1000)",
                },
            ],
        },
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".[0]",
        "payload": {"$where": "sleep(1000)"},
    }


def test_safe_email_password(create_context):
    assert (
        detect_nosql_injection(
            create_context(
                body={
                    "email": "email",
                    "password": "password",
                }
            ),
            {
                "email": "email",
                "password": "password",
            },
        )
        == {}
    )


def test_flags_pipeline_aggregations(create_context):
    assert detect_nosql_injection(
        create_context(
            body=[
                {
                    "$lookup": {
                        "from": "users",
                        "localField": "Dummy-IdontExist",
                        "foreignField": "Dummy-IdontExist",
                        "as": "user_docs",
                    },
                },
                {
                    "$limit": 1,
                },
            ]
        ),
        [
            {
                "$lookup": {
                    "from": "users",
                    "localField": "Dummy-IdontExist",
                    "foreignField": "Dummy-IdontExist",
                    "as": "user_docs",
                },
            },
            {
                "$limit": 1,
            },
        ],
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".[0]",
        "payload": {
            "$lookup": {
                "from": "users",
                "localField": "Dummy-IdontExist",
                "foreignField": "Dummy-IdontExist",
                "as": "user_docs",
            },
        },
    }

    assert detect_nosql_injection(
        create_context(
            body={
                "username": {
                    "$gt": "",
                },
            }
        ),
        [
            {
                "$match": {
                    "username": {
                        "$gt": "",
                    },
                },
            },
            {
                "$group": {
                    "_id": "$username",
                    "count": {"$sum": 1},
                },
            },
        ],
    ) == {
        "injection": True,
        "source": "body",
        "pathToPayload": ".username",
        "payload": {
            "$gt": "",
        },
    }


def test_ignores_safe_pipeline_aggregations(create_context):
    assert (
        detect_nosql_injection(
            create_context(
                body={
                    "username": "admin",
                }
            ),
            [
                {
                    "$match": {
                        "username": "admin",
                    },
                },
                {
                    "$group": {
                        "_id": "$username",
                        "count": {"$sum": 1},
                    },
                },
            ],
        )
        == {}
    )



"""
Mainly exports inspect_getaddrinfo_result function
"""

from aikido_zen.helpers.try_parse_url import try_parse_url
from aikido_zen.context import get_current_context
from aikido_zen.helpers.logging import logger
from aikido_zen.errors import AikidoSSRF
from aikido_zen.helpers.blocking_enabled import is_blocking_enabled
from aikido_zen.thread.thread_cache import get_cache
from .imds import resolves_to_imds_ip
from .is_private_ip import is_private_ip
from .find_hostname_in_context import find_hostname_in_context
from .extract_ip_array_from_results import extract_ip_array_from_results
from .is_redirect_to_private_ip import is_redirect_to_private_ip


#  gets called when the result of the DNS resolution has come in
def inspect_getaddrinfo_result(dns_results, hostname, port):
    """Inspect the results of a getaddrinfo() call"""
    if not hostname or try_parse_url(hostname) is not None:
        #  If the hostname is an IP address, we don't need to inspect it
        logger.debug("Hostname %s is actually an IP address, ignoring", hostname)
        return

    if not inspect_dns_results(dns_results, hostname):
        return

    context = get_current_context()
    if not context:
        return  # Context should be set to check user input.
    if get_cache() and get_cache().is_bypassed_ip(context.remote_address):
        # We check for bypassed ip's here since it is not checked for us
        # in run_vulnerability_scan due to the exception for SSRF (see above code)
        return

    # attack_findings is an object containing source, pathToPayload and payload.
    attack_findings = find_hostname_in_context(hostname, context, port)
    if not attack_findings:
        # Hostname/port not found in context, checking for redirects
        logger.debug("Redirects : %s", context.outgoing_req_redirects)
        attack_findings = is_redirect_to_private_ip(hostname, context, port)

    if attack_findings:
        return {
            "module": "socket",
            "operation": "socket.getaddrinfo",
            "kind": "ssrf",
            "source": attack_findings["source"],
            "path": attack_findings["pathToPayload"],
            "metadata": get_metadata_for_ssrf_attack(hostname, port),
            "payload": attack_findings["payload"],
        }


def get_metadata_for_ssrf_attack(hostname, port):
    """Returns metadata either with a port if it exists or just hostname"""
    if port:
        return {"hostname": hostname, "port": str(port)}
    return {"hostname": hostname}


def inspect_dns_results(dns_results, hostname):
    """
    Blocks stored SSRF attack that target IMDS IP addresses and returns True
    if a private_ip is present.
    This function gets called by inspect_getaddrinfo_result after parsing the hostname.
    """
    ip_addresses = extract_ip_array_from_results(dns_results)
    if resolves_to_imds_ip(ip_addresses, hostname):
        #  An attacker could have stored a hostname in a database that points to an IMDS IP address
        #  We don't check if the user input contains the hostname because there's no context
        if is_blocking_enabled():
            raise AikidoSSRF()

    private_ip = next((ip for ip in ip_addresses if is_private_ip(ip)), None)
    return private_ip



import pytest
from .create_interval import create_interval, interval_loop
from unittest.mock import Mock, call


def test_create_interval_schedules_function():
    # Arrange
    event_scheduler = Mock()
    function = Mock()
    args = (1, 2)
    interval_in_secs = 5

    # Act
    create_interval(event_scheduler, interval_in_secs, function, args)

    # Assert
    event_scheduler.enter.assert_called_once_with(
        interval_in_secs,
        1,
        interval_loop,
        (event_scheduler, interval_in_secs, function, args),
    )


def test_interval_loop_executes_function_and_reschedules():
    # Arrange
    event_scheduler = Mock()
    function = Mock()
    args = (1, 2)
    interval_in_secs = 5

    # Act
    interval_loop(event_scheduler, interval_in_secs, function, args)

    # Assert
    function.assert_called_once_with(*args)
    event_scheduler.enter.assert_called_once_with(
        interval_in_secs,
        1,
        interval_loop,
        (event_scheduler, interval_in_secs, function, args),
    )


def test_multiple_calls_to_create_interval():
    # Arrange
    event_scheduler = Mock()
    function = Mock()
    args = (1, 2)
    interval_in_secs = 5

    # Act
    create_interval(event_scheduler, interval_in_secs, function, args)
    create_interval(event_scheduler, interval_in_secs, function, args)

    # Assert
    assert event_scheduler.enter.call_count == 2
    assert event_scheduler.enter.call_args_list == [
        call(
            interval_in_secs,
            1,
            interval_loop,
            (event_scheduler, interval_in_secs, function, args),
        ),
        call(
            interval_in_secs,
            1,
            interval_loop,
            (event_scheduler, interval_in_secs, function, args),
        ),
    ]


if __name__ == "__main__":
    pytest.main()



import pytest
from .imds import is_imds_ip_address


# Assuming the is_imds_ip_address function is defined in the same file or imported from another module
def is_imds_ip_address(ip: str) -> bool:
    # This is a placeholder for the actual implementation
    # You should replace this with the actual function from your code
    return ip in ["169.254.169.254", "fd00:ec2::254"]


def test_returns_true_for_imds_ip_addresses():
    assert is_imds_ip_address("169.254.169.254") is True
    assert is_imds_ip_address("fd00:ec2::254") is True


def test_returns_false_for_non_imds_ip_addresses():
    assert is_imds_ip_address("1.2.3.4") is False
    assert is_imds_ip_address("example.com") is False



import pytest
import pytest
from .get_clean_stacktrace import get_clean_stacktrace


def test_get_clean_stacktrace_no_aikido():
    """Test that the stack trace does not include aikido frames."""

    def dummy_function():
        return get_clean_stacktrace()

    result = dummy_function()

    assert "/site-packages/aikido_zen/" not in result


def test_get_clean_stacktrace_with_aikido():
    """Test that the stack trace includes non-aikido frames."""

    def dummy_function():
        return get_clean_stacktrace()

    result = dummy_function()



import pytest
from unittest.mock import patch
import aikido_zen.sinks.psycopg
from aikido_zen.background_process.comms import reset_comms


@pytest.fixture
def database_conn():
    import psycopg

    return psycopg.connect(
        host="127.0.0.1", user="user", password="password", dbname="db"
    )


def test_cursor_execute(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        query = "SELECT * FROM dogs"
        cursor.execute(query)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == query
        assert called_with["args"][1] == "postgres"
        assert called_with["op"] == "psycopg.Cursor.execute"
        assert called_with["kind"] == "sql_injection"
        mock_run_vulnerability_scan.assert_called_once()

        cursor.fetchall()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()


def test_cursor_execute_parameterized(database_conn):
    reset_comms()
    cursor = database_conn.cursor()
    query = "SELECT * FROM dogs WHERE dog_name = %s"
    params = ("Fido",)
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor.execute(query, params)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == "SELECT * FROM dogs WHERE dog_name = %s"
        assert called_with["args"][1] == "postgres"
        assert called_with["op"] == "psycopg.Cursor.execute"
        assert called_with["kind"] == "sql_injection"
        mock_run_vulnerability_scan.assert_called_once()

    cursor.fetchall()
    database_conn.close()


def test_cursor_executemany(database_conn):
    reset_comms()
    cursor = database_conn.cursor()
    query = "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
    params = [("doggo1", False), ("Rex", False), ("Buddy", True)]

    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor.executemany(query, params)

        # Check the last call to run_vulnerability_scan
        called_with = mock_run_vulnerability_scan.call_args[1]
        assert (
            called_with["args"][0]
            == "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        )
        assert called_with["args"][1] == "postgres"
        assert called_with["op"] == "psycopg.Cursor.executemany"
        assert called_with["kind"] == "sql_injection"
        mock_run_vulnerability_scan.assert_called()

    database_conn.commit()
    database_conn.close()


def test_cursor_copy(database_conn):
    reset_comms()
    cursor = database_conn.cursor()
    query = "COPY dogs FROM STDIN"

    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor.copy(query)

        called_with = mock_run_vulnerability_scan.call_args[1]
        assert called_with["args"][0] == query
        assert called_with["args"][1] == "postgres"
        assert called_with["op"] == "psycopg.Cursor.copy"
        assert called_with["kind"] == "sql_injection"
        mock_run_vulnerability_scan.assert_called_once()

    database_conn.close()



import pytest
from .build_route_from_url import build_route_from_url
from .match_endpoints import match_endpoints


def sample_route_metadata(
    route="/posts/:number", method="POST", url="http://localhost:4000/posts/3"
):
    return {"method": method, "url": url, "route": route}


def test_invalid_url_and_no_route():
    assert match_endpoints(sample_route_metadata(route=None, url="abc"), []) == None


def test_no_url_and_no_route():
    assert match_endpoints(sample_route_metadata(route=None, url=None), []) == None


def test_no_method():
    assert match_endpoints(sample_route_metadata(method=None), []) == None


def test_it_returns_undefined_if_nothing_found():
    assert match_endpoints(sample_route_metadata(), []) == None


def test_it_returns_endpoint_based_on_route():
    endpoints = [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]
    assert match_endpoints(sample_route_metadata(), endpoints) == endpoints


def test_it_returns_endpoint_based_on_relative_url():
    assert match_endpoints(
        sample_route_metadata(url="/posts/3"),
        [
            {
                "method": "POST",
                "route": "/posts/:number",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
        ],
    ) == [
        {
            "method": "POST",
            "route": "/posts/:number",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]


def test_it_returns_endpoint_based_on_wildcard():
    assert match_endpoints(
        sample_route_metadata(route=None),
        [
            {
                "method": "*",
                "route": "/posts/*",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
        ],
    ) == [
        {
            "method": "*",
            "route": "/posts/*",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]


def test_it_returns_endpoint_based_on_wildcard_with_relative_url():
    assert match_endpoints(
        sample_route_metadata(route=None, url="/posts/3"),
        [
            {
                "method": "*",
                "route": "/posts/*",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
        ],
    ) == [
        {
            "method": "*",
            "route": "/posts/*",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]


def test_it_favors_more_specific_wildcard():
    assert match_endpoints(
        sample_route_metadata(
            url="http://localhost:4000/posts/3/comments/10", route=None
        ),
        [
            {
                "method": "*",
                "route": "/posts/*",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
            {
                "method": "*",
                "route": "/posts/*/comments/*",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
        ],
    ) == [
        {
            "method": "*",
            "route": "/posts/*/comments/*",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
        {
            "method": "*",
            "route": "/posts/*",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]


def test_it_matches_wildcard_route_with_specific_method():
    assert match_endpoints(
        sample_route_metadata(
            url="http://localhost:4000/posts/3/comments/10", route=None
        ),
        [
            {
                "method": "POST",
                "route": "/posts/*/comments/*",
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 10,
                    "windowSizeInMS": 1000,
                },
                "forceProtectionOff": False,
            },
        ],
    ) == [
        {
            "method": "POST",
            "route": "/posts/*/comments/*",
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 10,
                "windowSizeInMS": 1000,
            },
            "forceProtectionOff": False,
        },
    ]


def test_it_prefers_specific_route_over_wildcard():
    assert match_endpoints(
        sample_route_metadata(
            url="http://localhost:4000/api/coach", route="/api/coach"
        ),
        [
            {
                "method": "*",
                "route": "/api/*",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 20,
                    "windowSizeInMS": 60000,
                },
            },
            {
                "method": "POST",
                "route": "/api/coach",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 100,
                    "windowSizeInMS": 60000,
                },
            },
        ],
    ) == [
        {
            "method": "POST",
            "route": "/api/coach",
            "forceProtectionOff": False,
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 100,
                "windowSizeInMS": 60000,
            },
        },
        {
            "method": "*",
            "route": "/api/*",
            "forceProtectionOff": False,
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 20,
                "windowSizeInMS": 60000,
            },
        },
    ]


def test_it_prefers_specific_method_over_wildcard_first_case():
    result = match_endpoints(
        sample_route_metadata(
            url="http://localhost:4000/api/test", route="/api/test", method="POST"
        ),
        [
            {
                "method": "*",
                "route": "/api/test",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 20,
                    "windowSizeInMS": 60000,
                },
            },
            {
                "method": "POST",
                "route": "/api/test",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 100,
                    "windowSizeInMS": 60000,
                },
            },
        ],
    )

    expected = {
        "method": "POST",
        "route": "/api/test",
        "forceProtectionOff": False,
        "rateLimiting": {
            "enabled": True,
            "maxRequests": 100,
            "windowSizeInMS": 60000,
        },
    }

    assert result[0] == expected


def test_it_prefers_specific_method_over_wildcard_second_case():
    result = match_endpoints(
        sample_route_metadata(
            url="http://localhost:4000/api/test", route="/api/test", method="POST"
        ),
        [
            {
                "method": "POST",
                "route": "/api/test",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 100,
                    "windowSizeInMS": 60000,
                },
            },
            {
                "method": "*",
                "route": "/api/test",
                "forceProtectionOff": False,
                "rateLimiting": {
                    "enabled": True,
                    "maxRequests": 20,
                    "windowSizeInMS": 60000,
                },
            },
        ],
    )

    expected = {
        "method": "POST",
        "route": "/api/test",
        "forceProtectionOff": False,
        "rateLimiting": {
            "enabled": True,
            "maxRequests": 100,
            "windowSizeInMS": 60000,
        },
    }

    assert result[0] == expected



"""Exports get_ratelimited_endpoint, see function docstring"""


def get_ratelimited_endpoint(endpoints, route):
    """
    Filters through endpoints, tries to find an exact match otherwise tries to find the
    endpoint with the least amount of allowed requests.
    """
    if not endpoints:
        return None
    matches = []
    for endpoint in endpoints:
        if (
            endpoint.get("rateLimiting")
            and endpoint["rateLimiting"].get("enabled") is True
        ):
            if endpoint.get("route") == route:
                # Exact match, return the exact match
                return endpoint
            matches.append(endpoint)

    if len(matches) <= 0:
        return None

    matches.sort(key=get_endpoint_ratelimiting_rate)

    return matches[0]


def get_endpoint_ratelimiting_rate(endpoint):
    """
    Returns the rate at which a route gets ratelimited (max requests / ms)
    """
    max_reqs = endpoint["rateLimiting"].get("maxRequests")
    window_ms = endpoint["rateLimiting"].get("windowSizeInMS")
    return max_reqs / window_ms



from multiprocessing.forkserver import connect_to_new_process

import pytest
from unittest.mock import MagicMock
from .sync_data import process_sync_data
from aikido_zen.background_process.routes import Routes
from aikido_zen.helpers.iplist import IPList
from ...storage.hostnames import Hostnames
from ...storage.statistics import Statistics


@pytest.fixture
def setup_connection_manager():
    """Fixture to set up a mock connection manager."""
    connection_manager = MagicMock()
    connection_manager.routes = Routes()
    connection_manager.hostnames = Hostnames()
    connection_manager.conf.endpoints = ["endpoint1", "endpoint2"]

    connection_manager.conf.bypassed_ips = IPList()
    connection_manager.conf.bypassed_ips.add("192.168.1.1")
    connection_manager.conf.blocked_uids = ["user1", "user2"]
    connection_manager.conf.last_updated_at = 200
    connection_manager.statistics = Statistics()
    connection_manager.middleware_installed = False
    return connection_manager


def test_process_sync_data_initialization(setup_connection_manager):
    """Test the initialization of routes and hits update."""
    connection_manager = setup_connection_manager
    test_hostnames = Hostnames()
    test_hostnames.add("example2.com", 443, 15)
    test_hostnames.add("bumblebee.com", 8080)

    data = {
        "current_routes": {
            "route1": {
                "method": "GET",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 5,
                "apispec": {"info": "API spec for resource"},
            },
            "route2": {
                "method": "POST",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 3,
                "apispec": {"info": "API spec for resource"},
            },
        },
        "stats": {
            "startedAt": 1,
            "endedAt": 1,
            "requests": {
                "total": 10,
                "aborted": 0,
                "attacksDetected": {
                    "total": 5,
                    "blocked": 0,
                },
            },
        },
        "middleware_installed": False,
        "hostnames": test_hostnames.as_array(),
    }

    result = process_sync_data(connection_manager, data, None)

    # Check that routes were initialized correctly
    assert len(connection_manager.routes) == 2
    assert (
        connection_manager.routes.get({"method": "GET", "route": "/api/v1/resource"})[
            "hits"
        ]
        == 5
    )
    assert (
        connection_manager.routes.get({"method": "POST", "route": "/api/v1/resource"})[
            "hits"
        ]
        == 3
    )

    # Check that the total requests were updated
    assert connection_manager.statistics.get_record()["requests"] == {
        "aborted": 0,
        "attacksDetected": {"blocked": 0, "total": 5},
        "total": 10,
    }

    # Check that the return value is correct
    assert result["routes"] == dict(connection_manager.routes.routes)
    assert result["config"] == connection_manager.conf
    assert connection_manager.middleware_installed == False
    assert connection_manager.hostnames.as_array() == [
        {"hits": 15, "hostname": "example2.com", "port": 443},
        {"hits": 1, "hostname": "bumblebee.com", "port": 8080},
    ]


def test_process_sync_data_with_last_updated_at_below_zero(setup_connection_manager):
    """Test the initialization of routes and hits update."""
    connection_manager = setup_connection_manager
    connection_manager.conf.last_updated_at = -1
    data = {
        "current_routes": {
            "route1": {
                "method": "GET",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 5,
                "apispec": {"info": "API spec for resource"},
            },
            "route2": {
                "method": "POST",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 3,
                "apispec": {"info": "API spec for resource"},
            },
        },
        "stats": {
            "startedAt": 1,
            "endedAt": 1,
            "requests": {
                "total": 10,
                "aborted": 0,
                "attacksDetected": {
                    "total": 5,
                    "blocked": 0,
                },
            },
        },
        "middleware_installed": True,
    }

    result = process_sync_data(connection_manager, data, None)

    # Check that routes were initialized correctly
    assert len(connection_manager.routes) == 2
    assert (
        connection_manager.routes.get({"method": "GET", "route": "/api/v1/resource"})[
            "hits"
        ]
        == 5
    )
    assert (
        connection_manager.routes.get({"method": "POST", "route": "/api/v1/resource"})[
            "hits"
        ]
        == 3
    )

    # Check that the total requests were updated
    assert connection_manager.statistics.get_record()["requests"] == {
        "aborted": 0,
        "attacksDetected": {"blocked": 0, "total": 5},
        "total": 10,
    }
    assert connection_manager.middleware_installed == True
    assert len(connection_manager.hostnames.as_array()) == 0
    # Check that the return value is correct
    assert result == {}


def test_process_sync_data_existing_route_and_hostnames(setup_connection_manager):
    """Test updating an existing route's hit count."""
    connection_manager = setup_connection_manager
    connection_manager.hostnames.add("example.com", 443, 200)
    connection_manager.hostnames.add("example.org", 443)
    connection_manager.hostnames.add("a.com", 443)

    hostnames_sync = Hostnames()
    hostnames_sync.add("example.com", 443, 15)
    hostnames_sync.add("c.com", 443)

    data = {
        "current_routes": {
            "route1": {
                "method": "GET",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 5,
                "apispec": {"info": "API spec for resource"},
            }
        },
        "stats": {
            "startedAt": 1,
            "endedAt": 1,
            "requests": {
                "total": 5,
                "aborted": 0,
                "attacksDetected": {
                    "total": 5,
                    "blocked": 0,
                },
            },
        },
        "hostnames": hostnames_sync.as_array(),
    }

    # First call to initialize the route
    process_sync_data(connection_manager, data, None)

    # Second call to update the existing route
    data_update = {
        "current_routes": {
            "route1": {
                "method": "GET",
                "path": "/api/v1/resource",
                "hits_delta_since_sync": 10,
                "apispec": {"info": "Updated API spec for resource"},
            }
        },
        "stats": {
            "startedAt": 1,
            "endedAt": 1,
            "requests": {
                "total": 15,
                "aborted": 0,
                "attacksDetected": {
                    "total": 5,
                    "blocked": 0,
                },
            },
        },
    }

    result = process_sync_data(connection_manager, data_update, None)

    # Check that the hit count was updated correctly
    assert (
        connection_manager.routes.get({"method": "GET", "route": "/api/v1/resource"})[
            "hits"
        ]
        == 15
    )

    # Check that the total requests were updated
    assert connection_manager.statistics.get_record()["requests"] == {
        "aborted": 0,
        "attacksDetected": {"blocked": 0, "total": 10},
        "total": 20,
    }
    assert connection_manager.middleware_installed == False
    assert connection_manager.hostnames.as_array() == [
        {"hits": 215, "hostname": "example.com", "port": 443},
        {"hits": 1, "hostname": "example.org", "port": 443},
        {"hits": 1, "hostname": "a.com", "port": 443},
        {"hits": 1, "hostname": "c.com", "port": 443},
    ]

    # Check that the return value is correct
    assert result["routes"] == dict(connection_manager.routes.routes)


def test_process_sync_data_no_routes(setup_connection_manager):
    """Test behavior when no routes are provided."""
    connection_manager = setup_connection_manager
    data = {"current_routes": {}, "reqs": 0}  # No requests to add

    result = process_sync_data(connection_manager, data, None)

    # Check that no routes were initialized
    assert len(connection_manager.routes) == 0

    # Check that the total requests remain unchanged



from aikido_zen.helpers.extract_data_from_xml_body import (
    extract_data_from_xml_body,
)
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, after, patch_function


@after
def _fromstring(func, instance, args, kwargs, return_value):
    text = get_argument(args, kwargs, 0, "text")
    register_call("lxml.etree.fromstring", "deserialize_op")

    if text:
        extract_data_from_xml_body(user_input=text, root_element=return_value)


@after
def _fromstringlist(func, instance, args, kwargs, return_value):
    strings = get_argument(args, kwargs, 0, "strings")
    register_call("lxml.etree.fromstringlist", "deserialize_op")

    for text in strings:
        extract_data_from_xml_body(user_input=text, root_element=return_value)


@on_import("lxml.etree", "lxml")
def patch(m):
    """
    patching module lxml.etree
    - patches function fromstring(text, ...)
    - patches function fromstringlist(strings, ...)
    (github src: https://github.com/lxml/lxml/blob/fe271a4b5a32e6e54d10983683f2f32b0647209a/src/lxml/etree.pyx#L3411)
    """
    patch_function(m, "fromstring", _fromstring)
    patch_function(m, "fromstringlist", _fromstringlist)



"""
Helper function file, see function docstring
"""

import socket


def get_ip():
    """Tries to fetch the IP and returns x.x.x.x on failure"""
    try:
        return socket.gethostbyname(socket.gethostname())
    except Exception:
        return "x.x.x.x"



import pytest
from .iplist import IPList, get_ip_address_type


def test_get_ip_address_type():
    # Valid IPv4 addresses
    assert get_ip_address_type("192.168.1.1") == "ipv4"
    assert get_ip_address_type("255.255.255.255") == "ipv4"
    assert get_ip_address_type("0.0.0.0") == "ipv4"

    # Valid IPv6 addresses
    assert get_ip_address_type("::1") == "ipv6"
    assert get_ip_address_type("2001:0db8:85a3:0000:0000:8a2e:0370:7334") == "ipv6"
    assert get_ip_address_type("::") == "ipv6"

    # Invalid IP addresses
    assert get_ip_address_type("invalid_ip") is None
    assert get_ip_address_type("") is None
    assert get_ip_address_type("192.168.1.256") is None  # Invalid octet
    assert (
        get_ip_address_type("2001:0db8:85a3:0000:0000:8a2e:0370:7334:1234") is None
    )  # Too many segments


def test_blocklist_add_address():
    blocklist = IPList()
    assert blocklist.matches("192.168.1.1") is False
    blocklist.add_address("192.168.1.1", "ipv4")
    assert blocklist.matches("192.168.1.1") is True

    # Test adding the same address again
    blocklist.add_address("192.168.1.1", "ipv4")
    assert len(blocklist.blocked_addresses) == 1  # Should still be one entry


def test_blocklist_add_subnet():
    blocklist = IPList()
    assert blocklist.matches("10.0.0.1") is False
    blocklist.add_subnet("10.0.0.0", 8, "ipv4")
    assert blocklist.matches("10.0.0.1") is True
    assert blocklist.matches("10.1.1.1") is True
    assert blocklist.matches("192.168.1.1") is False


def test_blocklist_multiple_addresses_and_subnets():
    blocklist = IPList()
    blocklist.add_address("192.168.1.1", "ipv4")
    blocklist.add_subnet("10.0.0.0", 8, "ipv4")

    assert blocklist.matches("192.168.1.1") is True
    assert blocklist.matches("10.0.0.1") is True
    assert blocklist.matches("10.1.1.1") is True
    assert blocklist.matches("172.16.0.1") is False


def test_blocklist_invalid_ip():
    blocklist = IPList()
    blocklist.add_address("192.168.1.1", "ipv4")
    assert blocklist.matches("invalid_ip") is False


def test_blocklist_subnet_edge_cases():
    blocklist = IPList()
    blocklist.add_subnet("192.168.1.0", 24, "ipv4")

    assert blocklist.matches("192.168.1.255") is True  # Last address in the subnet
    assert blocklist.matches("192.168.1.0") is True  # First address in the subnet
    assert blocklist.matches("192.168.2.1") is False  # Outside the subnet
    assert blocklist.matches("192.168.1.128") is True  # Middle of the subnet


def test_blocklist_ipv6():
    blocklist = IPList()
    blocklist.add_address("2001:0db8:85a3:0000:0000:8a2e:0370:7334", "ipv6")
    assert blocklist.matches("2001:0db8:85a3:0000:0000:8a2e:0370:7334") is True
    assert blocklist.matches("::1") is False  # Not in the blocklist

    blocklist.add_subnet("2001:0db8::", 32, "ipv6")
    assert blocklist.matches("2001:0db8:abcd:0012:0000:0000:0000:0001") is True
    assert blocklist.matches("2001:0db9::") is False  # Outside the subnet


def test_blocklist_overlapping_subnets():
    blocklist = IPList()
    blocklist.add_subnet(
        "192.168.1.0", 24, "ipv4"
    )  # Covers 192.168.1.0 to 192.168.1.255
    blocklist.add_subnet(
        "192.168.1.128", 25, "ipv4"
    )  # Covers 192.168.1.128 to 192.168.1.255

    assert blocklist.matches("192.168.1.130") is True  # Inside both subnets
    assert blocklist.matches("192.168.1.0") is True  # Inside first subnet
    assert (
        blocklist.matches("192.168.1.127") is True
    )  # Inside first subnet, outside second
    assert blocklist.matches("192.168.1.255") is True  # Last address in both subnets
    assert blocklist.matches("192.168.2.1") is False  # Outside both subnets


def test_blocklist_mixed_address_types():
    blocklist = IPList()
    blocklist.add_address("192.168.1.1", "ipv4")
    blocklist.add_address("2001:0db8:85a3:0000:0000:8a2e:0370:7334", "ipv6")

    assert blocklist.matches("192.168.1.1") is True
    assert blocklist.matches("2001:0db8:85a3:0000:0000:8a2e:0370:7334") is True
    assert blocklist.matches("192.168.1.2") is False  # Different IPv4
    assert (
        blocklist.matches("2001:0db8:85a3:0000:0000:8a2e:0370:7335") is False
    )  # Different IPv6


def test_blocklist_subnet_with_single_ip():
    blocklist = IPList()
    blocklist.add_subnet("192.168.1.1", 32, "ipv4")  # Single IP subnet

    assert blocklist.matches("192.168.1.1") is True
    assert blocklist.matches("192.168.1.2") is False  # Outside the subnet


# Tests add function
def test_valid_ips():
    blocklist = IPList()

    blocklist.add("1.2.3.4")
    assert blocklist.matches("1.2.3.4") is True

    blocklist.add("fd00:ec2::254")
    assert blocklist.matches("fd00:ec2::254") is True

    blocklist.add("192.168.2.1/24")
    assert blocklist.matches("192.168.2.1") is True
    assert blocklist.matches("192.168.2.240") is True

    blocklist.add("fd00:124::1/64")
    assert blocklist.matches("fd00:124::1") is True
    assert blocklist.matches("fd00:124::f") is True
    assert blocklist.matches("fd00:124::ff13") is True

    blocklist.add("fd00:f123::1/128")
    assert blocklist.matches("fd00:f123::1") is True

    assert blocklist.matches("2.3.4.5") is False
    assert blocklist.matches("fd00:125::ff13") is False
    assert blocklist.matches("fd00:f123::2") is False


def test_invalid_ips():
    blocklist = IPList()

    blocklist.add("192.168.2.2.1/24")
    assert blocklist.matches("192.168.2.2.1") is False

    blocklist.add("test")
    assert blocklist.matches("test") is False

    blocklist.add("")
    assert blocklist.matches("") is False

    blocklist.add("fd00:124::1/test")
    assert blocklist.matches("fd00:124::1") is False



import pytest
import pickle
import json
from aikido_zen.context import Context, get_current_context, current_context


basic_wsgi_req = {
    "REQUEST_METHOD": "GET",
    "HTTP_HEADER_1": "header 1 value",
    "HTTP_HEADER_2": "Header 2 value",
    "RANDOM_VALUE": "Random value",
    "HTTP_COOKIE": "sessionId=abc123xyz456;",
    "wsgi.url_scheme": "http",
    "HTTP_HOST": "localhost:8080",
    "PATH_INFO": "/hello",
    "QUERY_STRING": "user=JohnDoe&age=30&age=35",
    "CONTENT_TYPE": "application/json",
    "REMOTE_ADDR": "198.51.100.23",
    "HTTP_USER_AGENT": "Mozilla/5.0",
}


@pytest.fixture(autouse=True)
def run_around_tests():
    yield
    # Make sure to reset context after every test so it does not
    # interfere with other tests
    current_context.set(None)


def test_get_current_context_no_context():
    # Test get_current_context() when no context is set
    assert get_current_context() is None


def test_wsgi_context_1():
    wsgi_request = {
        "REQUEST_METHOD": "POST",
        "HTTP_HEADER_1": "header 1 value",
        "HTTP_HEADER_2": "Header 2 value",
        "RANDOM_VALUE": "Random value",
        "HTTP_COOKIE": "sessionId=abc123xyz456;",
        "wsgi.url_scheme": "https",
        "HTTP_HOST": "example.com",
        "PATH_INFO": "/hello",
        "QUERY_STRING": "user=JohnDoe&age=30&age=35",
        "CONTENT_TYPE": "application/x-www-form-urlencoded",
        "REMOTE_ADDR": "198.51.100.23",
    }
    context = Context(req=wsgi_request, body=123, source="django")
    assert context.__dict__ == {
        "source": "django",
        "method": "POST",
        "headers": {
            "HEADER_1": "header 1 value",
            "HEADER_2": "Header 2 value",
            "COOKIE": "sessionId=abc123xyz456;",
            "HOST": "example.com",
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
        },
        "cookies": {"sessionId": "abc123xyz456"},
        "url": "https://example.com/hello",
        "query": {"user": ["JohnDoe"], "age": ["30", "35"]},
        "body": 123,
        "route": "/hello",
        "subdomains": [],
        "user": None,
        "remote_address": "198.51.100.23",
        "parsed_userinput": {},
        "xml": {},
        "outgoing_req_redirects": [],
        "executed_middleware": False,
        "route_params": [],
    }
    assert context.get_user_agent() is None


def test_wsgi_context_2():
    context = Context(req=basic_wsgi_req, body={"test": True}, source="flask")
    assert context.__dict__ == {
        "source": "flask",
        "method": "GET",
        "headers": {
            "HEADER_1": "header 1 value",
            "HEADER_2": "Header 2 value",
            "COOKIE": "sessionId=abc123xyz456;",
            "HOST": "localhost:8080",
            "CONTENT_TYPE": "application/json",
            "USER_AGENT": "Mozilla/5.0",
        },
        "cookies": {"sessionId": "abc123xyz456"},
        "url": "http://localhost:8080/hello",
        "query": {"user": ["JohnDoe"], "age": ["30", "35"]},
        "body": {"test": True},
        "route": "/hello",
        "subdomains": [],
        "user": None,
        "remote_address": "198.51.100.23",
        "parsed_userinput": {},
        "xml": {},
        "outgoing_req_redirects": [],
        "executed_middleware": False,
        "route_params": [],
    }
    assert context.get_user_agent() == "Mozilla/5.0"


def test_set_as_current_context(mocker):
    # Test set_as_current_context() method
    context = Context(req=basic_wsgi_req, body=12, source="flask")
    context.set_as_current_context()
    assert get_current_context() == context


def test_get_current_context_with_context(mocker):
    # Test get_current_context() when a context is set
    context = Context(req=basic_wsgi_req, body=456, source="flask")
    context.set_as_current_context()
    assert get_current_context() == context


def test_context_is_picklable(mocker):
    context = Context(req=basic_wsgi_req, body=123, source="flask")
    pickled_obj = pickle.dumps(context)
    unpickled_obj = pickle.loads(pickled_obj)
    assert unpickled_obj.source == "flask"
    assert unpickled_obj.method == "GET"
    assert unpickled_obj.remote_address == "198.51.100.23"
    assert unpickled_obj.url == "http://localhost:8080/hello"
    assert unpickled_obj.body == 123
    assert unpickled_obj.headers == {
        "HEADER_1": "header 1 value",
        "HEADER_2": "Header 2 value",
        "COOKIE": "sessionId=abc123xyz456;",
        "HOST": "localhost:8080",
        "CONTENT_TYPE": "application/json",
        "USER_AGENT": "Mozilla/5.0",
    }
    assert unpickled_obj.query == {"user": ["JohnDoe"], "age": ["30", "35"]}
    assert unpickled_obj.cookies == {"sessionId": "abc123xyz456"}


def test_set_valid_dict():
    valid_body = {"key": "value"}
    context = Context(req=basic_wsgi_req, body=valid_body, source="flask")
    assert context.body == valid_body


def test_set_valid_list():
    valid_body = [1, 2, 3]
    context = Context(req=basic_wsgi_req, body=valid_body, source="flask")
    assert context.body == valid_body


def test_set_valid_string():
    valid_body = "This is a valid string"
    context = Context(req=basic_wsgi_req, body=valid_body, source="flask")
    assert context.body == valid_body


def test_set_empty_byte_string():
    invalid_body = b""
    context = Context(req=basic_wsgi_req, body=invalid_body, source="flask")
    assert context.body is None


def test_set_normal_byte_string():
    body = b"Hello World!"
    context = Context(req=basic_wsgi_req, body=body, source="flask")
    assert context.body == "Hello World!"


def test_set_byte_string_wrong_encoding():
    body = "hello world! 😊".encode("utf-16")  # UTF-16 unique character
    context = Context(req=basic_wsgi_req, body=body, source="flask")
    assert context.body == body  # Body remains unchanged because utf-8 failed.


def test_set_none():
    context = Context(req=basic_wsgi_req, body=None, source="flask")
    assert context.body is None


def test_set_valid_nested_json_string():
    context = Context(req=basic_wsgi_req, body=None, source="flask")
    context.set_body('{"key": {"nested_key": "nested_value"}}')
    assert context.body == {"key": {"nested_key": "nested_value"}}


def test_set_invalid_json_with_unmatched_quotes():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('{"key": "value\'s}')
    assert context.body == '{"key": "value\'s}'  # Should remain as string


def test_set_valid_json_with_array():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('{"key": [1, 2, 3]}')
    assert context.body == {"key": [1, 2, 3]}


def test_set_valid_json_with_spaces():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('               {"key": [1, 2, 3]}         ')
    assert context.body == {"key": [1, 2, 3]}


def test_valid_json_string_with_newlines():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('\r\n\r\n"hello"\r\n\r\n')
    assert context.body == "hello"


def test_valid_json_string_with_spaces():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('"                    hello              "')
    assert context.body == "                    hello              "


def test_set_valid_json_with_newlines():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body('\r\n\r\n{"key": [1, 2, 3]}\r\n\r\n')
    assert context.body == {"key": [1, 2, 3]}


def test_set_valid_json_with_spaces_and_array():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body("               [1, 2, 3]         ")
    assert context.body == [1, 2, 3]
    context.set_body("               (1, 2, 3)         ")
    assert context.body == "               (1, 2, 3)         "


def test_set_valid_json_with_spaces_and_array_bytes():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body(b"               [1, 2, 3]         ")
    assert context.body == [1, 2, 3]
    context.set_body(b"               (1, 2, 3)         ")
    assert context.body == "               (1, 2, 3)         "


def test_set_valid_json_with_complex_array():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body(
        '               [{"hello": "world", "one": 123}, 2, "hiya"]         '
    )
    assert context.body == [{"hello": "world", "one": 123}, 2, "hiya"]


def test_set_valid_json_with_complex_array_bytes():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body(
        b'               [{"hello": "world", "one": 123}, 2, "hiya"]         '
    )
    assert context.body == [{"hello": "world", "one": 123}, 2, "hiya"]


def test_empty_string_becomes_none():
    context = Context(req=basic_wsgi_req, body=None, source="flask")

    context.set_body("")
    assert context.body is None


def test_set_valid_json_with_special_characters():
    context = Context(req=basic_wsgi_req, body=None, source="flask")
    context.set_body('{"key": "value with special characters !@#$%^&*()"}')
    assert context.body == {"key": "value with special characters !@#$%^&*()"}


def test_set_valid_json_with_special_characters_bytes():
    context = Context(req=basic_wsgi_req, body=None, source="flask")
    context.set_body(b'{"key": "value with special characters !@#$%^&*()"}')
    assert context.body == {"key": "value with special characters !@#$%^&*()"}



import pytest
from .get_ip_from_request import (
    get_ip_from_request,
    is_ip,
    get_client_ip_from_x_forwarded_for,
)


# Test `get_ip_from_request` function :
def test_get_ip_from_request():
    # Test case 1: Valid X_FORWARDED_FOR header with valid IP
    headers = {"X_FORWARDED_FOR": "192.168.1.1, 10.0.0.1"}
    assert get_ip_from_request(None, headers) == "192.168.1.1"

    # Test case 2: Valid X_FORWARDED_FOR header with invalid IPs
    headers = {"X_FORWARDED_FOR": "256.256.256.256, 192.168.1.1"}
    assert (
        get_ip_from_request(None, headers) == "192.168.1.1"
    )  # Should return the valid IP

    # Test case 3: Valid remote address
    headers = {}
    assert get_ip_from_request("10.0.0.1", headers) == "10.0.0.1"

    # Test case 4: Valid remote address with invalid X_FORWARDED_FOR
    headers = {"X_FORWARDED_FOR": "abc.def.ghi.jkl, 256.256.256.256"}
    assert (
        get_ip_from_request("10.0.0.1", headers) == "10.0.0.1"
    )  # Should return the remote address

    # Test case 5: Both X_FORWARDED_FOR and remote address are invalid
    headers = {"X_FORWARDED_FOR": "abc.def.ghi.jkl, 256.256.256.256"}
    assert get_ip_from_request(None, headers) is None  # Should return None

    # Test case 6: Empty headers and remote address
    headers = {}
    assert get_ip_from_request(None, headers) is None  # Should return None


#  Test `is_ip` function :
def test_valid_ipv4():
    assert is_ip("192.168.1.1")  # Valid IPv4
    assert is_ip("255.255.255.255")  # Valid IPv4
    assert is_ip("0.0.0.0")  # Valid IPv4


def test_invalid_ipv4():
    assert not is_ip("256.256.256.256")  # Invalid IPv4
    assert not is_ip("192.168.1")  # Invalid IPv4
    assert not is_ip("abc.def.ghi.jkl")  # Invalid IPv4


def test_valid_ipv6():
    assert is_ip("::1")  # Valid IPv6 (loopback)
    assert is_ip("2001:0db8:85a3:0000:0000:8a2e:0370:7334")  # Valid IPv6


def test_invalid_ipv6():
    assert not is_ip("2001:db8:85a3::8a2e:370:7334:12345")  # Invalid IPv6
    assert not is_ip("::g")  # Invalid IPv6


#  Test `get_client_ip_from_x_forwarded_for` function :
def test_get_client_ip_from_x_forwarded_for():
    # Test cases with valid IPs
    assert get_client_ip_from_x_forwarded_for("192.168.1.1") == "192.168.1.1"
    assert get_client_ip_from_x_forwarded_for("192.168.1.1, 10.0.0.1") == "192.168.1.1"
    assert get_client_ip_from_x_forwarded_for("10.0.0.1, 192.168.1.1") == "10.0.0.1"
    assert (
        get_client_ip_from_x_forwarded_for("2001:0db8:85a3:0000:0000:8a2e:0370:7334")
        == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    )
    assert (
        get_client_ip_from_x_forwarded_for(
            "2001:0db8:85a3:0000:0000:8a2e:0370:7334, 192.168.1.1"
        )
        == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    )

    # Test cases with mixed valid and invalid IPs
    assert (
        get_client_ip_from_x_forwarded_for("256.256.256.256, 192.168.1.1")
        == "192.168.1.1"
    )  # Invalid IPv4 ignored
    assert (
        get_client_ip_from_x_forwarded_for("192.168.1.1, abc.def.ghi.jkl")
        == "192.168.1.1"
    )  # Invalid IPv4 ignored
    assert (
        get_client_ip_from_x_forwarded_for(
            "::1, 2001:0db8:85a3:0000:0000:8a2e:0370:7334"
        )
        == "::1"
    )  # Valid IPv6 preferred

    # Test cases with only invalid IPs
    assert (
        get_client_ip_from_x_forwarded_for("abc.def.ghi.jkl, 256.256.256.256") is None
    )  # All invalid
    assert get_client_ip_from_x_forwarded_for("") is None  # Empty string



import pytest
from unittest.mock import MagicMock, patch
from aikido_zen.helpers.token import Token
from .start_polling_for_changes import start_polling_for_changes, poll_for_changes


@pytest.fixture
def event_scheduler():
    class EventScheduler:
        def __init__(self):
            self.events = []

        def enter(self, delay, priority, action, argument):
            self.events.append((delay, priority, action, argument))

        def run(self):
            if self.events:
                self.events[0][2](*self.events[0][3])

    return EventScheduler()


def test_no_token(event_scheduler, caplog):
    start_polling_for_changes(
        connection_manager=MagicMock(token=None, serverless=None),
        event_scheduler=event_scheduler,
    )

    assert "No token provided, not polling for config updates" in caplog.text


def test_serverless_environment(event_scheduler, caplog):
    start_polling_for_changes(
        connection_manager=MagicMock(token=Token("123"), serverless=True),
        event_scheduler=event_scheduler,
    )

    assert (
        "Running in serverless environment, not polling for config updates"
        in caplog.text
    )


def test_check_for_config_updates(event_scheduler):
    calls = []
    config_updated_at = 0

    def mock_get_config_last_updated_at(token):
        return config_updated_at

    def mock_get_config(token):
        return {"endpoints": [], "heartbeatIntervalInMS": 600000}

    with patch(
        "aikido_zen.background_process.realtime.get_config_last_updated_at",
        side_effect=mock_get_config_last_updated_at,
    ), patch(
        "aikido_zen.background_process.realtime.get_config",
        side_effect=mock_get_config,
    ):
        config_updates = []

        connection_manager = MagicMock(
            update_service_config=lambda config: config_updates.append(config),
            token=Token("123"),
            conf=MagicMock(last_updated_at=config_updated_at),
            serverless=None,
        )

        start_polling_for_changes(
            connection_manager=connection_manager,
            event_scheduler=event_scheduler,
        )

        # Run the event scheduler to simulate the polling
        event_scheduler.run()

        assert config_updates == []

        # Simulate a config update
        config_updated_at = 1
        connection_manager.conf.last_updated_at = config_updated_at
        event_scheduler.run()

        assert config_updates == [
            {
                "endpoints": [],
                "heartbeatIntervalInMS": 600000,
                "success": True,
            },
        ]


def test_api_error_handling(event_scheduler, caplog):
    def mock_get_config_last_updated_at(token):
        raise ValueError("Request timed out")

    with patch(
        "aikido_zen.background_process.realtime.get_config_last_updated_at",
        side_effect=mock_get_config_last_updated_at,
    ):
        config_updates = []

        connection_manager = MagicMock(
            update_service_config=lambda config: config_updates.append(config),
            token=Token("123"),
            conf=MagicMock(last_updated_at=0),
            serverless=None,
        )

        start_polling_for_changes(
            connection_manager=connection_manager,
            event_scheduler=event_scheduler,
        )

        # Run the event scheduler to simulate the polling
        event_scheduler.run()

        assert config_updates == []
        assert (
            "Failed to check for config updates due to error : Request timed out"
            in caplog.text
        )



"""
SQL Injection algorithm
"""

import re
import ctypes
from aikido_zen.helpers.logging import logger
from .map_dialect_to_rust_int import map_dialect_to_rust_int
from .get_lib_path import get_binary_path


def detect_sql_injection(query, user_input, dialect):
    """
    Execute this to check if the query is actually a SQL injection
    """
    try:
        query_l = query.lower()
        userinput_l = user_input.lower()
        if should_return_early(query_l, userinput_l):
            return False

        internals_lib = ctypes.CDLL(get_binary_path())
        query_bytes = query_l.encode("utf-8")
        userinput_bytes = userinput_l.encode("utf-8")
        dialect_int = map_dialect_to_rust_int(dialect)
        c_int_res = internals_lib.detect_sql_injection(
            query_bytes, userinput_bytes, dialect_int
        )

        # This means that an error occurred in the library
        if c_int_res == 2:
            logger.debug(
                "Unable to check for SQL Injection, an error occurred in the library"
            )
            return False

        return bool(c_int_res)
    except Exception as e:
        logger.debug("Exception in SQL algo: %s", e)
    return False


def should_return_early(query, user_input):
    """
    Input : Lowercased query and user_input.
    Returns true if the detect_sql_injection algo should return early :
    - user_input is <= 1 char or user input larger than query
    - user_input not in query
    - user_input is alphanumerical
    - user_input is an array of integers
    """
    if len(user_input) <= 1 or len(query) < len(user_input):
        # User input too small or larger than query, returning
        return True

    if user_input not in query:
        # User input not in query, returning
        return True

    if user_input.replace("_", "").isalnum():
        # User input is alphanumerical (with underscores), returning
        return True

    cleaned_input_for_list = user_input.replace(" ", "").replace(",", "")
    is_valid_comma_seperated_number_list = re.match(r"^\d+$", cleaned_input_for_list)
    if is_valid_comma_seperated_number_list:
        # E.g. 1, 24028, 828, 4, 1
        return True

    return False



import pytest
from unittest.mock import patch

from aikido_zen.background_process.cloud_connection_manager import (
    CloudConnectionManager,
)
from .update_firewall_lists import update_firewall_lists
from ...helpers.is_ip_allowed_by_allowlist import is_ip_allowed_by_allowlist


class MockApi:
    """Mock API class to simulate API responses."""

    def fetch_firewall_lists(self, token):
        return {
            "success": True,
            "blockedIPAddresses": [
                {
                    "source": "example",
                    "description": "Example description",
                    "ips": ["192.168.1.0/24"],
                }
            ],
            "allowedIPAddresses": [
                {
                    "source": "list1",
                    "description": "Example description",
                    "ips": ["192.168.1.3"],
                },
                {
                    "source": "example",
                    "description": "Example description",
                    "ips": ["192.168.2.0/24"],
                },
            ],
            "blockedUserAgents": "BadBot|Test",
        }


class MockApiNoUA:
    """Mock API class to simulate API responses."""

    def fetch_firewall_lists(self, token):
        return {
            "success": True,
            "blockedIPAddresses": [
                {
                    "source": "example",
                    "description": "Example description",
                    "ips": ["192.168.1.0/24"],
                }
            ],
            "allowedIPAddresses": [],
            "blockedUserAgents": "",
        }


class MockApiInvalidRegex:
    """Mock API class to simulate API responses."""

    def fetch_firewall_lists(self, token):
        return {
            "success": True,
            "blockedIPAddresses": [
                {
                    "source": "example",
                    "description": "Example description",
                    "ips": ["192.168.1.0/24"],
                }
            ],
            "allowedIPAddresses": [],
            "blockedUserAgents": "[abc",
        }


@pytest.fixture
def connection_manager():
    """Fixture to create a CloudConnectionManager instance with a mock API."""

    class TestCloudConnectionManager(CloudConnectionManager):
        """A test version of CloudConnectionManager with a mock API."""

        def __init__(self, token, serverless):
            super().__init__(
                block=False, api=MockApi(), token=token, serverless=serverless
            )

    return TestCloudConnectionManager(token="valid_token", serverless=False)


def test_update_firewall_lists_success(connection_manager):
    # Call the function to test
    update_firewall_lists(connection_manager)

    # Check that the blocked IPs were set correctly
    assert connection_manager.conf.is_blocked_ip("192.168.1.1")
    assert connection_manager.conf.is_blocked_ip("192.168.1.2")

    # Check that the allowed IPs were set correctly
    assert is_ip_allowed_by_allowlist(connection_manager.conf, "192.168.1.3")
    assert is_ip_allowed_by_allowlist(connection_manager.conf, "192.168.2.50")

    # Check that the blocked user agents were set correctly
    assert connection_manager.conf.is_user_agent_blocked("bAdBoT test woop wop")
    assert not connection_manager.conf.is_user_agent_blocked("")
    assert not connection_manager.conf.is_user_agent_blocked(None)


def test_update_firewall_lists_no_ua(connection_manager):
    # Call the function to test
    connection_manager.api = MockApiNoUA()
    update_firewall_lists(connection_manager)

    # Check that the blocked IPs were set correctly
    assert connection_manager.conf.is_blocked_ip("192.168.1.1")
    assert connection_manager.conf.is_blocked_ip("192.168.1.2")

    # Check that the blocked user agents were set correctly
    assert not connection_manager.conf.is_user_agent_blocked("bAdBoT test woop wop")
    assert not connection_manager.conf.is_user_agent_blocked("")
    assert connection_manager.conf.blocked_user_agent_regex is None


def test_update_firewall_lists_invalid_regex(connection_manager):
    # Call the function to test
    connection_manager.api = MockApiInvalidRegex()
    update_firewall_lists(connection_manager)

    # Check that the blocked IPs were set correctly
    assert connection_manager.conf.is_blocked_ip("192.168.1.1")
    assert connection_manager.conf.is_blocked_ip("192.168.1.2")

    # Check that the blocked user agents were set correctly
    assert not connection_manager.conf.is_user_agent_blocked("bAdBoT test woop wop")
    assert not connection_manager.conf.is_user_agent_blocked("")
    assert connection_manager.conf.blocked_user_agent_regex is None


def test_update_firewall_lists_no_token(connection_manager):
    # Set token to None
    connection_manager.token = None
    update_firewall_lists(connection_manager)
    # No changes should be made, so we check the initial state
    assert len(connection_manager.conf.bypassed_ips.blocked_subnets) == 0
    assert len(connection_manager.conf.bypassed_ips.blocked_addresses) == 0


def test_update_firewall_lists_serverless(connection_manager):
    # Set serverless to True
    connection_manager.serverless = True
    update_firewall_lists(connection_manager)
    # No changes should be made, so we check the initial state
    assert len(connection_manager.conf.bypassed_ips.blocked_subnets) == 0
    assert len(connection_manager.conf.bypassed_ips.blocked_addresses) == 0


def test_update_firewall_lists_api_failure(connection_manager):
    # Override the mock API to simulate a failure
    class FailingApi:
        def fetch_firewall_lists(self, token):
            return {"success": False}

    connection_manager.api = FailingApi()
    update_firewall_lists(connection_manager)
    # No changes should be made, so we check the initial state
    assert len(connection_manager.conf.bypassed_ips.blocked_subnets) == 0
    assert len(connection_manager.conf.bypassed_ips.blocked_addresses) == 0


def test_update_firewall_lists_exception_handling(connection_manager):
    # Override the mock API to raise an exception
    class ExceptionApi:
        def fetch_firewall_lists(self, token):
            raise Exception("API error")

    connection_manager.api = ExceptionApi()
    update_firewall_lists(connection_manager)
    # No changes should be made, so we check the initial state
    assert len(connection_manager.conf.bypassed_ips.blocked_subnets) == 0
    assert len(connection_manager.conf.bypassed_ips.blocked_addresses) == 0



import pytest
from .ai_statistics import AIStatistics


@pytest.fixture
def stats():
    return AIStatistics()


def test_initializes_with_empty_state(stats):
    assert stats.get_stats() == []
    assert stats.empty() is True


def test_tracks_basic_ai_calls(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    result = stats.get_stats()
    assert len(result) == 1
    assert result[0] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 1,
        "tokens": {
            "input": 100,
            "output": 50,
            "total": 150,
        },
    }

    assert stats.empty() is False


def test_tracks_multiple_calls_to_same_provider_model(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=200, output_tokens=75
    )

    result = stats.get_stats()
    assert len(result) == 1
    assert result[0] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 2,
        "tokens": {
            "input": 300,
            "output": 125,
            "total": 425,
        },
    }


def test_tracks_different_provider_model_combinations_separately(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    stats.on_ai_call(
        provider="openai", model="gpt-3.5-turbo", input_tokens=80, output_tokens=40
    )

    stats.on_ai_call(
        provider="anthropic", model="claude-3", input_tokens=120, output_tokens=60
    )

    result = stats.get_stats()
    assert len(result) == 3

    # Sort by provider:model for consistent testing
    result.sort(key=lambda x: f"{x['provider']}:{x['model']}")

    assert result[0] == {
        "provider": "anthropic",
        "model": "claude-3",
        "calls": 1,
        "tokens": {
            "input": 120,
            "output": 60,
            "total": 180,
        },
    }

    assert result[1] == {
        "provider": "openai",
        "model": "gpt-3.5-turbo",
        "calls": 1,
        "tokens": {
            "input": 80,
            "output": 40,
            "total": 120,
        },
    }

    assert result[2] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 1,
        "tokens": {
            "input": 100,
            "output": 50,
            "total": 150,
        },
    }


def test_resets_all_statistics(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    stats.on_ai_call(
        provider="anthropic", model="claude-3", input_tokens=120, output_tokens=60
    )

    assert stats.empty() is False
    assert len(stats.get_stats()) == 2

    stats.clear()

    assert stats.empty() is True
    assert stats.get_stats() == []


def test_handles_zero_token_inputs(stats):
    stats.on_ai_call(provider="openai", model="gpt-4", input_tokens=0, output_tokens=0)

    result = stats.get_stats()
    assert len(result) == 1
    assert result[0]["tokens"] == {
        "input": 0,
        "output": 0,
        "total": 0,
    }


def test_called_with_empty_provider(stats):
    stats.on_ai_call(provider="", model="gpt-4", input_tokens=100, output_tokens=50)

    assert stats.empty() is True


def test_called_with_empty_model(stats):
    stats.on_ai_call(provider="openai", model="", input_tokens=100, output_tokens=50)

    assert stats.empty() is True


def test_get_stats_returns_immutable_data(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    result = stats.get_stats()
    result[0]["calls"] = 100
    result[0]["tokens"]["input"] = 1000

    # Verify that the internal state has not changed
    assert stats.get_stats()[0]["calls"] == 1


def test_get_stats_returns_new_list(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    result1 = stats.get_stats()
    result2 = stats.get_stats()

    # Modify the first result to ensure it doesn't affect the second result
    result1[0]["calls"] = 200

    # Verify that the second result is unchanged
    assert result2 == [
        {
            "provider": "openai",
            "model": "gpt-4",
            "calls": 1,
            "tokens": {
                "input": 100,
                "output": 50,
                "total": 150,
            },
        }
    ]


def test_get_stats_returns_deep_copy(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    result = stats.get_stats()

    # Modify the result deeply to ensure it doesn't affect the internal state
    result[0]["tokens"]["input"] = 200

    # Verify that the internal state has not changed
    assert stats.get_stats()[0]["tokens"]["input"] == 100


def test_get_stats_consistency_after_multiple_calls(stats):
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    stats.on_ai_call(
        provider="anthropic", model="claude-3", input_tokens=120, output_tokens=60
    )

    result1 = stats.get_stats()
    result2 = stats.get_stats()

    # Modify the first result to ensure it doesn't affect the second result
    result1[0]["calls"] = 300

    # Verify that the second result is unchanged
    assert result2 == [
        {
            "provider": "openai",
            "model": "gpt-4",
            "calls": 1,
            "tokens": {
                "input": 100,
                "output": 50,
                "total": 150,
            },
        },
        {
            "provider": "anthropic",
            "model": "claude-3",
            "calls": 1,
            "tokens": {
                "input": 120,
                "output": 60,
                "total": 180,
            },
        },
    ]


def test_import_list_with_empty_stats(stats):
    # Test importing a list into empty statistics
    stats.import_list(
        [
            {
                "provider": "openai",
                "model": "gpt-4",
                "calls": 1,
                "tokens": {
                    "input": 100,
                    "output": 50,
                    "total": 150,
                },
            },
            {
                "provider": "anthropic",
                "model": "claude-3",
                "calls": 1,
                "tokens": {
                    "input": 120,
                    "output": 60,
                    "total": 180,
                },
            },
        ]
    )

    result = stats.get_stats()
    assert len(result) == 2
    assert result[0] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 1,
        "tokens": {
            "input": 100,
            "output": 50,
            "total": 150,
        },
    }
    assert result[1] == {
        "provider": "anthropic",
        "model": "claude-3",
        "calls": 1,
        "tokens": {
            "input": 120,
            "output": 60,
            "total": 180,
        },
    }


def test_import_list_with_existing_stats(stats):
    # Add some initial statistics
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )

    # Import a list that includes an existing provider:model combination
    stats.import_list(
        [
            {
                "provider": "openai",
                "model": "gpt-4",
                "calls": 1,
                "tokens": {
                    "input": 200,
                    "output": 100,
                    "total": 300,
                },
            },
            {
                "provider": "anthropic",
                "model": "claude-3",
                "calls": 1,
                "tokens": {
                    "input": 120,
                    "output": 60,
                    "total": 180,
                },
            },
        ]
    )

    result = stats.get_stats()
    assert len(result) == 2
    assert result[0] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 2,  # Initial call + imported call
        "tokens": {
            "input": 300,  # 100 + 200
            "output": 150,  # 50 + 100
            "total": 450,  # 150 + 300
        },
    }
    assert result[1] == {
        "provider": "anthropic",
        "model": "claude-3",
        "calls": 1,
        "tokens": {
            "input": 120,
            "output": 60,
            "total": 180,
        },
    }


def test_import_list_with_overlapping_and_new_entries(stats):
    # Add some initial statistics
    stats.on_ai_call(
        provider="openai", model="gpt-4", input_tokens=100, output_tokens=50
    )
    stats.on_ai_call(
        provider="anthropic", model="claude-3", input_tokens=120, output_tokens=60
    )

    # Import a list that includes both existing and new provider:model combinations
    stats.import_list(
        [
            {
                "provider": "openai",
                "model": "gpt-4",
                "calls": 1,
                "tokens": {
                    "input": 200,
                    "output": 100,
                    "total": 300,
                },
            },
            {
                "provider": "anthropic",
                "model": "claude-3",
                "calls": 1,
                "tokens": {
                    "input": 120,
                    "output": 60,
                    "total": 180,
                },
            },
            {
                "provider": "mistral",
                "model": "mistral-7b",
                "calls": 1,
                "tokens": {
                    "input": 150,
                    "output": 75,
                    "total": 225,
                },
            },
        ]
    )

    result = stats.get_stats()
    assert len(result) == 3
    result.sort(key=lambda x: f"{x['provider']}:{x['model']}")

    assert result[0] == {
        "provider": "anthropic",
        "model": "claude-3",
        "calls": 2,  # Initial call + imported call
        "tokens": {
            "input": 240,  # 120 + 120
            "output": 120,  # 60 + 60
            "total": 360,  # 180 + 180
        },
    }
    assert result[1] == {
        "provider": "mistral",
        "model": "mistral-7b",
        "calls": 1,
        "tokens": {
            "input": 150,
            "output": 75,
            "total": 225,
        },
    }
    assert result[2] == {
        "provider": "openai",
        "model": "gpt-4",
        "calls": 2,  # Initial call + imported call
        "tokens": {
            "input": 300,  # 100 + 200
            "output": 150,  # 50 + 100
            "total": 450,  # 150 + 300
        },
    }



import pytest
from .try_parse_url import try_parse_url


def test_valid_http_url():
    url = "http://example.com"
    result = try_parse_url(url)
    assert result is not None
    assert result.scheme == "http"
    assert result.netloc == "example.com"


def test_valid_https_url():
    url = "https://example.com/path?query=1"
    result = try_parse_url(url)
    assert result is not None
    assert result.scheme == "https"
    assert result.netloc == "example.com"
    assert result.path == "/path"
    assert result.query == "query=1"


def test_invalid_url_missing_scheme():
    url = "example.com/path"
    result = try_parse_url(url)
    assert result is None


def test_invalid_url_missing_netloc():
    url = "http:///path"
    result = try_parse_url(url)
    assert result is None


def test_invalid_url():
    url = "ht!tp://example.com"
    result = try_parse_url(url)
    assert result is None


def test_empty_url():
    url = ""
    result = try_parse_url(url)
    assert result is None


def test_url_with_only_scheme():
    url = "http://"
    result = try_parse_url(url)
    assert result is None


def test_url_with_special_characters():
    url = "http://example.com/path?query=1&other=value#fragment"
    result = try_parse_url(url)
    assert result is not None
    assert result.scheme == "http"
    assert result.netloc == "example.com"
    assert result.path == "/path"
    assert result.query == "query=1&other=value"
    assert result.fragment == "fragment"



"""Exports extract_route_params function"""

from urllib.parse import quote, unquote
from aikido_zen.helpers.try_parse_url_path import try_parse_url_path
from aikido_zen.helpers.build_route_from_url import replace_url_segment_with_param


def extract_route_params(url):
    """Will try and build an array of user input based on the url"""
    results = []
    try:
        path = try_parse_url_path(url)
        segments = path.split("/")
        for segment in segments:
            segment = unquote(segment)
            if segment.isalnum():
                continue  # Ignore alphanumerical parts of the url

            if segment is not quote(segment):
                results.append(segment)  # This is not a standard piece of the URL
            elif replace_url_segment_with_param(segment) is not segment:
                results.append(segment)  # Might be a secret, a hash, ...

        if len(results) > 0 or "." in unquote(path):
            # There are already phishy parts of the url OR
            # urldecoded path contains dots, which is uncommon and could point to path traversal.
            results.append(path[1:])  # Add path after slash as user input

    except Exception:
        pass
    return results



import sys

import pytest
from unittest.mock import patch
import aikido_zen.sinks.psycopg2
from aikido_zen.background_process.comms import reset_comms

kind = "sql_injection"
op = "pymysql.connections.query"

# psycopg2 not working on python 3.13 for now
skip_python_3_13 = sys.version_info[:2] == (3, 13)


@pytest.fixture
def database_conn():
    import psycopg2
    from psycopg2.extras import DictCursor

    return psycopg2.connect(
        cursor_factory=DictCursor,
        host="127.0.0.1",
        user="user",
        password="password",
        database="db",
    )


@pytest.fixture
def database_conn_immutable_cursor():
    import psycopg2
    from psycopg2.extensions import cursor

    return psycopg2.connect(
        cursor_factory=cursor,
        host="127.0.0.1",
        user="user",
        password="password",
        database="db",
    )


@pytest.fixture
def database_conn_empty_cursor():
    import psycopg2

    return psycopg2.connect(
        host="127.0.0.1",
        user="user",
        password="password",
        database="db",
    )


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_execute(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        print(cursor)
        query = "SELECT * FROM dogs"
        cursor.execute(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        cursor.fetchall()
        database_conn.close()
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_execute2(database_conn_empty_cursor):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn_empty_cursor.cursor()
        query = "SELECT * FROM dogs"
        cursor.execute(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        cursor.fetchall()
        database_conn_empty_cursor.close()
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_execute3(database_conn_immutable_cursor):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn_immutable_cursor.cursor()
        query = "SELECT * FROM dogs"
        cursor.execute(query)

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        cursor.fetchall()
        database_conn_immutable_cursor.close()
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_execute_parameterized(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        query = "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        cursor.execute(query, ("doggo", True))

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        database_conn.commit()
        database_conn.close()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_execute_parameterized2(database_conn_empty_cursor):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn_empty_cursor.cursor()
        query = "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        cursor.execute(query, ("doggo", True))

        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert called_with_args[0] == query
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        database_conn_empty_cursor.commit()
        database_conn_empty_cursor.close()
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_executemany(database_conn):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn.cursor()
        data = [
            ("Doggy", False),
            ("Doggy 2", True),
            ("Dogski", True),
        ]
        cursor.executemany("INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)", data)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0]
            == "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        )
        assert called_with_args[1] == "postgres"

        database_conn.commit()
        cursor.close()
        database_conn.close()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_executemany2(database_conn_empty_cursor):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn_empty_cursor.cursor()
        data = [
            ("Doggy", False),
            ("Doggy 2", True),
            ("Dogski", True),
        ]
        cursor.executemany("INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)", data)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0]
            == "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        )
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        database_conn_empty_cursor.commit()
        cursor.close()
        database_conn_empty_cursor.close()
        mock_run_vulnerability_scan.assert_called_once()


@pytest.mark.skipif(skip_python_3_13, reason="Skipping test on Python 3.13")
def test_cursor_executemany3(database_conn_immutable_cursor):
    reset_comms()
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        cursor = database_conn_immutable_cursor.cursor()
        data = [
            ("Doggy", False),
            ("Doggy 2", True),
            ("Dogski", True),
        ]
        cursor.executemany("INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)", data)
        called_with_args = mock_run_vulnerability_scan.call_args[1]["args"]
        assert (
            called_with_args[0]
            == "INSERT INTO dogs (dog_name, isadmin) VALUES (%s, %s)"
        )
        assert called_with_args[1] == "postgres"
        mock_run_vulnerability_scan.assert_called_once()

        database_conn_immutable_cursor.commit()
        cursor.close()
        database_conn_immutable_cursor.close()
        mock_run_vulnerability_scan.assert_called_once()



"""Helper function file, see function docstring"""

import json


def serialize_to_json(obj):
    """
    Tries to serialize the object, if that fails
    it tries to convert the object to a dict and then serialize
    if that fails it returns an empty JSON string "{}"
    """
    try:
        return json.dumps(obj)
    except Exception:
        try:
            return json.dumps(dict(obj))
        except Exception:
            pass
    return "{}"



from .is_primitive_type import only_contains_primitive_types, is_primitive_type


def test_only_contains_primitive_types_single_primitive():
    assert only_contains_primitive_types("string") is True
    assert only_contains_primitive_types("number") is True
    assert only_contains_primitive_types("boolean") is True


def test_only_contains_primitive_types_single_non_primitive():
    assert only_contains_primitive_types("object") is False
    assert only_contains_primitive_types("array") is False


def test_only_contains_primitive_types_multiple_primitives():
    assert only_contains_primitive_types(["string", "number", "boolean"]) is True


def test_only_contains_primitive_types_mixed():
    assert only_contains_primitive_types(["string", "object", "number"]) is False
    assert only_contains_primitive_types(["array", "boolean"]) is False


def test_only_contains_primitive_types_empty():
    assert only_contains_primitive_types([]) is True  # An empty list should return True


def test_only_contains_primitive_types_non_list_input():
    assert only_contains_primitive_types("string") is True
    assert only_contains_primitive_types("object") is False



"""
Helper module for token
"""

import os


class Token:
    """Class that encapsulates the token"""

    def __init__(self, token):
        if not isinstance(token, str):
            raise ValueError("Token should be an instance of string")
        if len(token) == 0:
            raise ValueError("Token cannot be an empty string")
        self.token = token

    def __str__(self):
        return self.token


def get_token_from_env():
    """
    Fetches the token from the env variable "AIKIDO_TOKEN"
    """
    aikido_token_env = os.getenv("AIKIDO_TOKEN")
    if aikido_token_env is not None:
        return Token(aikido_token_env)
    return None



"""
Sink module for `psycopg`
"""

import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import patch_function, on_import, before


@before
def _copy(func, instance, args, kwargs):
    statement = get_argument(args, kwargs, 0, "statement")

    op = "psycopg.Cursor.copy"
    register_call(op, "sql_op")

    vulns.run_vulnerability_scan(
        kind="sql_injection", op=op, args=(statement, "postgres")
    )


@before
def _execute(func, instance, args, kwargs):
    query = get_argument(args, kwargs, 0, "query")
    op = f"psycopg.Cursor.{func.__name__}"
    vulns.run_vulnerability_scan(kind="sql_injection", op=op, args=(query, "postgres"))


@on_import("psycopg.cursor", "psycopg", version_requirement="3.1.0")
def patch(m):
    """
    patching module psycopg.cursor
    - patches Cursor.copy
    - patches Cursor.execute
    - patches Cursor.executemany
    """
    patch_function(m, "Cursor.copy", _copy)
    patch_function(m, "Cursor.execute", _execute)
    patch_function(m, "Cursor.executemany", _execute)



import pytest
from .parse_cookies import parse_cookies


def test_parse_cookies_single_cookie():
    cookie_str = "sessionid=abc123"
    expected = {"sessionid": "abc123"}
    assert parse_cookies(cookie_str) == expected


def test_parse_cookies_multiple_cookies():
    cookie_str = "sessionid=abc123; user=JohnDoe; theme=dark"
    expected = {"sessionid": "abc123", "user": "JohnDoe", "theme": "dark"}
    assert parse_cookies(cookie_str) == expected


def test_parse_cookies_empty_string():
    cookie_str = ""
    expected = {}
    assert parse_cookies(cookie_str) == expected


def test_parse_cookies_with_special_characters():
    cookie_str = "sessionid=abc123; user=John@Doe; theme=dark&light"
    expected = {"sessionid": "abc123", "user": "John@Doe", "theme": "dark&light"}
    assert parse_cookies(cookie_str) == expected



from aikido_zen.helpers.logging import logger
from .extract_data_from_request import extract_data_from_request
from ..functions.request_handler import request_handler
from ...sinks import on_import, patch_function, before


def _request_response(func, instance, args, kwargs):
    if kwargs and "func" in kwargs:
        kwargs["func"] = aik_route_func_wrapper(kwargs["func"])
    elif args and args[0]:
        # Modify first element of a tuple, tuples are immutable
        args = (aik_route_func_wrapper(args[0]),) + args[1:]

    return func(*args, **kwargs)  # Call the original function


@on_import("starlette.routing", "starlette")
def patch(m):
    """
    patching module starlette.routing (for initial request_handler)
    - patches: request_response
    (github src: https://github.com/encode/starlette/blob/4acf1d1ca3e8aa767567cb4e6e12f093f066553b/starlette/routing.py#L58)
    """
    patch_function(m, "request_response", _request_response)


def aik_route_func_wrapper(func):
    async def aikido_route_func(*args, **kwargs):
        # Code before response (pre_response stage)
        try:
            req = args[0]
            if not req:
                return
            await extract_data_from_request(req)
            pre_response_results = request_handler(stage="pre_response")
            if pre_response_results:
                response = create_starlette_response(pre_response_results)
                if response:
                    # Make sure to not return when an error occurred or there is an invalid response
                    return response
        except Exception as e:
            logger.debug("Exception occurred in pre_response stage starlette : %s", e)

        # Calling the function, check if it's async or not (same checks as in codebase starlette)
        try:
            import functools
            from starlette.concurrency import run_in_threadpool
            from starlette._utils import is_async_callable
        except ImportError:
            logger.info("Make sure starlette install OK : .concurrency, ._utils")
            return await func(*args, **kwargs)
        res = None
        if is_async_callable(func):
            res = await func(*args, **kwargs)
        else:
            # `func` provided by the end-user is allowed to be both synchronous and asynchronous
            # Here we convert sync functions in the same way the starlette codebase does to ensure
            # there are no compatibility issues and the behaviour remains unchanged.
            res = await functools.partial(run_in_threadpool, func)(*args, **kwargs)

        # Code after response (post_response stage)
        request_handler(stage="post_response", status_code=res.status_code)
        return res

    return aikido_route_func


def create_starlette_response(pre_response):
    """Tries to import PlainTextResponse and generates starlette plain text response"""
    text, status_code = pre_response
    try:
        from starlette.responses import PlainTextResponse
    except ImportError:
        logger.info(
            "Ensure `starlette` install is valid, failed to import starlette.responses"
        )
        return None
    return PlainTextResponse(text, status_code)



"""
Sink module for python's `os`
"""

from pathlib import PurePath
import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import before, patch_function, on_import


@before
def _os_patch(func, instance, args, kwargs):
    possible_paths = args + tuple(kwargs.values())
    for path in possible_paths:
        if not isinstance(path, (str, bytes, PurePath)):
            continue
        # change op if it's an os.path function
        op = f"os.{func.__name__}"
        if func.__name__ in ("getsize", "join", "expanduser", "expandvars", "realpath"):
            op = f"os.path.{func.__name__}"
        register_call(op, "fs_op")

        vulns.run_vulnerability_scan(kind="path_traversal", op=op, args=(path,))


@on_import("os")
def patch(m):
    """
    patching module os
    - patches os.* functions that take in paths
    - patches os.path.* functions that take in paths
    - doesn't patch os.makedirs -> uses os.mkdir
    - doesn't patch os.path.abspath -> uses os.path.join
    - doesn't patch os.path.exists and others -> to big of a performance impact
    - doesn't patch stat library https://docs.python.org/3/library/stat.html
    """
    # os.*(...) patches
    patch_function(m, "access", _os_patch)
    patch_function(m, "chmod", _os_patch)
    patch_function(m, "chown", _os_patch)
    patch_function(m, "mkdir", _os_patch)
    patch_function(m, "listdir", _os_patch)
    patch_function(m, "readlink", _os_patch)
    patch_function(m, "unlink", _os_patch)
    patch_function(m, "rename", _os_patch)
    patch_function(m, "rmdir", _os_patch)
    patch_function(m, "remove", _os_patch)
    patch_function(m, "symlink", _os_patch)
    patch_function(m, "link", _os_patch)
    patch_function(m, "walk", _os_patch)
    patch_function(m, "open", _os_patch)

    # os.path.*(...) patches
    patch_function(m, "path.getsize", _os_patch)
    patch_function(m, "path.join", _os_patch)
    patch_function(m, "path.expanduser", _os_patch)
    patch_function(m, "path.expandvars", _os_patch)
    patch_function(m, "path.realpath", _os_patch)  # Python 3.13



from aikido_zen.helpers.build_route_from_url import build_route_from_url
import pytest
import hashlib


def generate_hash(algorithm):
    data = "test"
    if algorithm == "md5":
        return hashlib.md5(data.encode()).hexdigest()
    elif algorithm == "sha1":
        return hashlib.sha1(data.encode()).hexdigest()
    elif algorithm == "sha256":
        return hashlib.sha256(data.encode()).hexdigest()
    elif algorithm == "sha512":
        return hashlib.sha512(data.encode()).hexdigest()
    else:
        return None


def test_invalid_urls():
    assert build_route_from_url("") == None
    assert build_route_from_url("http") == None


def test_root_urls():
    assert build_route_from_url("/") == "/"
    assert build_route_from_url("http://localhost/") == "/"


def test_replace_numbers():
    assert build_route_from_url("/posts/3") == "/posts/:number"
    assert build_route_from_url("http://localhost/posts/3") == "/posts/:number"
    assert build_route_from_url("http://localhost/posts/3/") == "/posts/:number"
    assert (
        build_route_from_url("http://localhost/posts/3/comments/10")
        == "/posts/:number/comments/:number"
    )
    assert (
        build_route_from_url("/blog/2023/05/great-article")
        == "/blog/:number/:number/great-article"
    )


def test_replace_dates():
    assert build_route_from_url("/posts/2023-05-01") == "/posts/:date"
    assert build_route_from_url("/posts/2023-05-01/") == "/posts/:date"
    assert (
        build_route_from_url("/posts/2023-05-01/comments/2023-05-01")
        == "/posts/:date/comments/:date"
    )
    assert build_route_from_url("/posts/01-05-2023") == "/posts/:date"


def test_ignore_comma_numbers():
    assert build_route_from_url("/posts/3,000") == "/posts/3,000"


def test_ignore_api_version_numbers():
    assert build_route_from_url("/v1/posts/3") == "/v1/posts/:number"


def test_replace_uuids():
    uuids = [
        "d9428888-122b-11e1-b85c-61cd3cbb3210",
        "000003e8-2363-21ef-b200-325096b39f47",
        "a981a0c2-68b1-35dc-bcfc-296e52ab01ec",
        "109156be-c4fb-41ea-b1b4-efe1671c5836",
        "90123e1c-7512-523e-bb28-76fab9f2f73d",
        "1ef21d2f-1207-6660-8c4f-419efbd44d48",
        "017f22e2-79b0-7cc3-98c4-dc0c0c07398f",
        "0d8f23a0-697f-83ae-802e-48f3756dd581",
    ]
    for uuid in uuids:
        assert build_route_from_url(f"/posts/{uuid}") == "/posts/:uuid"


def test_ignore_invalid_uuids():
    assert (
        build_route_from_url("/posts/00000000-0000-1000-6000-000000000000")
        == "/posts/00000000-0000-1000-6000-000000000000"
    )


def test_ignore_strings():
    assert build_route_from_url("/posts/abc") == "/posts/abc"


def test_replace_email_addresses():
    assert build_route_from_url("/login/john.doe@acme.com") == "/login/:email"
    assert build_route_from_url("/login/john.doe+alias@acme.com") == "/login/:email"


def test_replace_ip_addresses():
    assert build_route_from_url("/block/1.2.3.4") == "/block/:ip"
    assert (
        build_route_from_url("/block/2001:2:ffff:ffff:ffff:ffff:ffff:ffff")
        == "/block/:ip"
    )
    assert build_route_from_url("/block/64:ff9a::255.255.255.255") == "/block/:ip"
    assert build_route_from_url("/block/100::") == "/block/:ip"
    assert build_route_from_url("/block/fec0::") == "/block/:ip"
    assert build_route_from_url("/block/227.202.96.196") == "/block/:ip"


def test_replace_hashes():
    assert build_route_from_url(f"/files/{generate_hash('md5')}") == "/files/:hash"
    assert build_route_from_url(f"/files/{generate_hash('sha1')}") == "/files/:hash"
    assert build_route_from_url(f"/files/{generate_hash('sha256')}") == "/files/:hash"
    assert build_route_from_url(f"/files/{generate_hash('sha512')}") == "/files/:hash"


def test_replace_secrets():
    assert (
        build_route_from_url("/confirm/CnJ4DunhYfv2db6T1FRfciRBHtlNKOYrjoz")
        == "/confirm/:secret"
    )



from unittest.mock import patch, MagicMock

import pytest
from aikido_zen.context import current_context, Context, get_current_context
from aikido_zen.thread.thread_cache import ThreadCache, get_cache
from . import should_block_request


@pytest.fixture(autouse=True)
def run_around_tests():
    get_cache().reset()
    yield
    # Make sure to reset context and cache after every test so it does not
    # interfere with other tests
    current_context.set(None)
    get_cache().reset()


def test_without_context():
    current_context.set(None)
    assert should_block_request() == {"block": False}


def set_context(user=None, executed_middleware=False):
    Context(
        context_obj={
            "remote_address": "::1",
            "method": "POST",
            "url": "http://localhost:4000",
            "query": {
                "abc": "def",
            },
            "headers": {},
            "body": None,
            "cookies": {},
            "source": "flask",
            "route": "/posts/:id",
            "user": user,
            "executed_middleware": executed_middleware,
        }
    ).set_as_current_context()


def test_with_context_without_cache():
    set_context()
    get_cache().cache = None
    assert should_block_request() == {"block": False}


def test_with_context_with_cache():
    set_context(user={"id": "123"})
    thread_cache = get_cache()

    thread_cache.config.blocked_uids = ["123"]
    assert get_current_context().executed_middleware == False
    assert thread_cache.middleware_installed == False
    assert should_block_request() == {
        "block": True,
        "trigger": "user",
        "type": "blocked",
    }
    assert get_current_context().executed_middleware == True
    assert thread_cache.middleware_installed == True

    thread_cache.config.blocked_uids = []
    assert should_block_request() == {"block": False}

    thread_cache.config.blocked_uids = ["23", "234", "456"]
    assert should_block_request() == {"block": False}
    assert get_current_context().executed_middleware == True
    assert thread_cache.middleware_installed == True


def test_cache_comms_with_endpoints():
    set_context(user={"id": "456"})
    thread_cache = get_cache()
    thread_cache.config.blocked_uids = ["123"]
    thread_cache.config.endpoints = [
        {
            "method": "POST",
            "route": "/login",
            "forceProtectionOff": False,
            "rateLimiting": {
                "enabled": True,
                "maxRequests": 3,
                "windowSizeInMS": 1000,
            },
        }
    ]
    assert get_current_context().executed_middleware == False
    assert thread_cache.middleware_installed == False

    with patch("aikido_zen.background_process.comms.get_comms") as mock_get_comms:
        mock_get_comms.return_value = None  # Set the return value of get_comms
        assert should_block_request() == {"block": False}
    assert get_current_context().executed_middleware == True
    assert thread_cache.middleware_installed == True

    with patch("aikido_zen.background_process.comms.get_comms") as mock_get_comms:
        mock_comms = MagicMock()
        mock_get_comms.return_value = mock_comms  # Set the return value of get_comms

        # No matching endpoints :
        assert should_block_request() == {"block": False}
        mock_comms.send_data_to_bg_process.assert_not_called()

    thread_cache.config.endpoints.append(
        {
            "method": "POST",
            "route": "/posts/:id",
            "forceProtectionOff": False,
            "rateLimiting": {
                "enabled": False,
                "maxRequests": 3,
                "windowSizeInMS": 1000,
            },
        }
    )

    with patch("aikido_zen.background_process.comms.get_comms") as mock_get_comms:
        mock_comms = MagicMock()
        mock_get_comms.return_value = mock_comms  # Set the return value of get_comms

        # Rate-limiting disabled:
        assert should_block_request() == {"block": False}
        mock_comms.send_data_to_bg_process.assert_not_called()

    # Enable ratelimiting
    thread_cache.config.endpoints[1]["rateLimiting"]["enabled"] = True

    with patch("aikido_zen.background_process.comms.get_comms") as mock_get_comms:
        mock_comms = MagicMock()
        mock_get_comms.return_value = mock_comms  # Set the return value of get_comms
        mock_comms.send_data_to_bg_process.return_value = {"success": False}

        assert should_block_request() == {"block": False}
        mock_comms.send_data_to_bg_process.assert_called_with(
            action="SHOULD_RATELIMIT",
            obj={
                "route_metadata": {
                    "method": "POST",
                    "route": "/posts/:id",
                    "url": "http://localhost:4000",
                },
                "user": {"id": "456"},
                "remote_address": "::1",
            },
            receive=True,
        )

        mock_comms.send_data_to_bg_process.return_value = {
            "success": True,
            "data": {"block": False, "trigger": "my_trigger"},
        }
        assert should_block_request() == {"block": False}

        mock_comms.send_data_to_bg_process.return_value = {
            "success": True,
            "data": {"block": True, "trigger": "my_trigger"},
        }
        assert should_block_request() == {
            "block": True,
            "ip": "::1",
            "type": "ratelimited",
            "trigger": "my_trigger",
        }



import pytest
from aikido_zen.background_process.comms import AikidoIPCCommunications


def test_comms_init():
    address = ("localhost", 9898)
    key = "secret_key"
    comms = AikidoIPCCommunications(address, key)

    assert comms.address == address
    assert comms.key == key


def test_send_data_to_bg_process_exception(monkeypatch, caplog):
    def mock_client(address, authkey):
        raise Exception("Connection Error")

    monkeypatch.setitem(globals(), "Client", mock_client)
    monkeypatch.setitem(globals(), "logger", caplog)

    comms = AikidoIPCCommunications(("localhost", 9898), "mock_key")
    comms.send_data_to_bg_process("ACTION", "Test Object")


def test_send_data_to_bg_process_successful(monkeypatch, caplog, mocker):
    comms = AikidoIPCCommunications(("localhost"), "mock_key")
    mock_client = mocker.MagicMock()
    monkeypatch.setattr("multiprocessing.connection.Client", mock_client)

    # Call the send_data_to_bg_process function
    comms.send_data_to_bg_process("ACTION", {"key": "value"})



import pytest
from unittest.mock import MagicMock, patch
from .find_hostname_in_context import (
    find_hostname_in_context,
)  # Replace with the actual module name


# Mocking the dependencies
def mock_extract_strings_from_user_input_cached(user_input, source):
    return {"example_input": "path/to/payload"}


def mock_find_hostname_in_userinput(user_input, hostname, port):
    return user_input == hostname


# Test cases
def test_find_hostname_in_context_found(monkeypatch):
    # Arrange
    hostname = "example.com"
    context = MagicMock()
    context.cookies = "example.com"
    context.body = "another user input"
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    with patch(
        "aikido_zen.helpers.extract_strings_from_user_input.extract_strings_from_user_input_cached",
        side_effect=mock_extract_strings_from_user_input_cached,
    ):
        with patch(
            "aikido_zen.vulnerabilities.ssrf.find_hostname_in_userinput",
            side_effect=mock_find_hostname_in_userinput,
        ):  # Replace 'your_module' with the actual module name
            # Act
            result = find_hostname_in_context(hostname, context, 80)

            # Assert
            assert result == {
                "source": "cookies",
                "pathToPayload": ".",
                "payload": "example.com",
            }


def test_find_hostname_in_context_punycode_hostname(monkeypatch):
    hostname = "app.xn--loca-3b5a.aikido.dev"
    context = MagicMock()
    context.body = "app.locaḷ.aikido.dev"
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    with patch(
        "aikido_zen.helpers.extract_strings_from_user_input.extract_strings_from_user_input_cached",
        side_effect=mock_extract_strings_from_user_input_cached,
    ):
        with patch(
            "aikido_zen.vulnerabilities.ssrf.find_hostname_in_userinput",
            side_effect=mock_find_hostname_in_userinput,
        ):
            result = find_hostname_in_context(hostname, context, 80)

            assert result == {
                "source": "body",
                "pathToPayload": ".",
                "payload": "app.locaḷ.aikido.dev",
            }


def test_find_hostname_in_context_not_found(monkeypatch):
    # Arrange
    hostname = "notfound.com"
    context = MagicMock()
    context.body = "some user input"
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    with patch(
        "aikido_zen.helpers.extract_strings_from_user_input.extract_strings_from_user_input_cached",
        side_effect=mock_extract_strings_from_user_input_cached,
    ):
        with patch(
            "aikido_zen.vulnerabilities.ssrf.find_hostname_in_userinput",
            side_effect=mock_find_hostname_in_userinput,
        ):  # Replace 'your_module' with the actual module name
            # Act
            result = find_hostname_in_context(hostname, context, 80)

            # Assert
            assert result is None


def test_find_hostname_in_context_no_sources(monkeypatch):
    # Arrange
    hostname = "example.com"
    context = MagicMock()  # No attributes
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    # Act
    result = find_hostname_in_context(hostname, context, 80)

    # Assert
    assert result is None


# To run the tests, use the command: pytest <filename>.py


@pytest.mark.parametrize(
    "invalid_input",
    [
        None,
        123456789,  # Integer
        45.67,  # Float
        [],  # Empty list
        [1, 2, 3],  # List of integers
        {},  # Empty dictionary
        {"key": "value"},  # Dictionary
        set(),  # Empty set
        {1, 2, 3},  # Set of integers
        object(),  # Instance of a generic object
        lambda x: x,  # Lambda function
        (1, 2),  # Tuple
        b"bytes",  # Bytes
    ],
)
def test_doesnt_crash_with_invalid_hostname(invalid_input, monkeypatch):
    context = MagicMock()  # No attributes
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)
    result = find_hostname_in_context(invalid_input, context, 8080)
    assert result == None


@pytest.mark.parametrize(
    "invalid_input",
    [
        None,
        "test",  # String
        45.67,  # Float
        [],  # Empty list
        [1, 2, 3],  # List of integers
        {},  # Empty dictionary
        {"key": "value"},  # Dictionary
        set(),  # Empty set
        {1, 2, 3},  # Set of integers
        object(),  # Instance of a generic object
        lambda x: x,  # Lambda function
        (1, 2),  # Tuple
        b"bytes",  # Bytes
    ],
)
def test_doesnt_crash_with_invalid_port(invalid_input, monkeypatch):
    context = MagicMock()  # No attributes
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)
    result = find_hostname_in_context("https://example.com", context, invalid_input)
    assert result == None



import aikido_zen.helpers.get_current_unixtime_ms as t
from aikido_zen.storage.statistics.operations import Operations


class Statistics:
    """
    Keeps track of total and aborted requests
    and total and blocked attacks
    """

    def __init__(self):
        self.total_hits = 0
        self.attacks_detected = 0
        self.attacks_blocked = 0
        self.started_at = t.get_unixtime_ms()
        self.operations = Operations()

    def clear(self):
        self.total_hits = 0
        self.attacks_detected = 0
        self.attacks_blocked = 0
        self.started_at = t.get_unixtime_ms()
        self.operations.clear()

    def increment_total_hits(self):
        self.total_hits += 1

    def on_detected_attack(self, blocked, operation):
        self.attacks_detected += 1
        if blocked:
            self.attacks_blocked += 1
        self.operations.on_detected_attack(blocked, operation)

    def get_record(self):
        current_time = t.get_unixtime_ms()
        return {
            "startedAt": self.started_at,
            "endedAt": current_time,
            "requests": {
                "total": self.total_hits,
                "aborted": 0,  # statistic currently not in use
                "attacksDetected": {
                    "total": self.attacks_detected,
                    "blocked": self.attacks_blocked,
                },
            },
            "operations": dict(self.operations),
        }

    def import_from_record(self, record):
        attacks_detected = record.get("requests", {}).get("attacksDetected", {})
        self.total_hits += record.get("requests", {}).get("total", 0)
        self.attacks_detected += attacks_detected.get("total", 0)
        self.attacks_blocked += attacks_detected.get("blocked", 0)
        self.operations.update(record.get("operations", {}))

    def empty(self):
        if self.total_hits > 0:
            return False
        if self.attacks_detected > 0:
            return False
        if len(self.operations) > 0:
            return False
        return True



"""Exports pre_response_middleware function"""

from ..functions.request_handler import request_handler


def pre_response_middleware(request, *args, **kwargs):
    """Aikido middleware that handles ratelimiting"""
    response = request_handler(stage="pre_response")
    if response:
        # pylint:disable=import-outside-toplevel # We don't want to install this by default
        from django.http import HttpResponse

        return HttpResponse(response[0], status=response[1])
    return None



import pytest
from .is_safely_encapsulated import is_safely_encapsulated


def test_safe_between_single_quotes():
    assert is_safely_encapsulated("echo '$USER'", "$USER") == True
    assert is_safely_encapsulated("echo '`$USER'", "`USER") == True


def test_single_quote_in_single_quotes():
    assert is_safely_encapsulated("echo ''USER'", "'USER") == False


def test_dangerous_chars_between_double_quotes():
    assert is_safely_encapsulated('echo "=USER"', "=USER") == True
    assert is_safely_encapsulated('echo "$USER"', "$USER") == False
    assert is_safely_encapsulated('echo "!USER"', "!USER") == False
    assert is_safely_encapsulated('echo "\\`USER"', "`USER") == False
    assert is_safely_encapsulated('echo "\\USER"', "\\USER") == False


def test_same_user_input_multiple_times():
    assert is_safely_encapsulated("echo '$USER' '$USER'", "$USER") == True
    assert is_safely_encapsulated("echo \"$USER\" '$USER'", "$USER") == False
    assert is_safely_encapsulated('echo "$USER" "$USER"', "$USER") == False


def test_first_and_last_quote_does_not_match():
    assert is_safely_encapsulated("echo '$USER\"", "$USER") == False
    assert is_safely_encapsulated("echo \"$USER'", "$USER") == False


def test_first_or_last_character_not_escape_char():
    assert is_safely_encapsulated("echo $USER'", "$USER") == False
    assert is_safely_encapsulated('echo $USER"', "$USER") == False


def test_user_input_does_not_occur_in_command():
    assert is_safely_encapsulated("echo 'USER'", "$USER") == True
    assert is_safely_encapsulated('echo "USER"', "$USER") == True



"""
Sink module for python's `shutil`
"""

import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, patch_function, before


@before
def _shutil_func(func, instance, args, kwargs):
    source = get_argument(args, kwargs, 0, "src")
    destination = get_argument(args, kwargs, 1, "dst")

    op = f"shutil.{func.__name__}"
    register_call(op, "fs_op")

    kind = "path_traversal"
    if isinstance(source, str):
        vulns.run_vulnerability_scan(kind, op, args=(source,))
    if isinstance(destination, str):
        vulns.run_vulnerability_scan(kind, op, args=(destination,))


@on_import("shutil")
def patch(m):
    """
    patching module shutil
    - patches: copymode, copystat, copytree, move
    - does not patch: copyfile, copy, copy2 -> uses builtins.open
    """
    patch_function(m, "copymode", _shutil_func)
    patch_function(m, "copystat", _shutil_func)
    patch_function(m, "copytree", _shutil_func)
    patch_function(m, "move", _shutil_func)



import pytest
from unittest.mock import patch
from aikido_zen.background_process.api.http_api import ReportingApiHTTP
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms
from .http_api_ratelimited import ReportingApiHTTPRatelimited


@pytest.fixture
def reporting_api():
    """Fixture to create an instance of ReportingApiHTTPRatelimited."""
    return ReportingApiHTTPRatelimited(
        reporting_url="http://example.com",
        max_events_per_interval=3,
        interval_in_ms=10000,
    )


def test_report_within_limit(reporting_api):
    """Test reporting within the rate limit."""
    event = {"type": "detected_attack", "time": 1000}

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        with patch(
            "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
            return_value=2000,
        ):
            response = reporting_api.report(
                token="token", event=event, timeout_in_sec=5
            )
            assert response == {"success": True}
            assert len(reporting_api.events) == 1
            mock_report.assert_called_once_with("token", event, 5)


def test_report_exceeds_limit(reporting_api):
    """Test reporting when the limit is exceeded."""
    event = {"type": "detected_attack", "time": 1000}

    # Simulate adding events to reach the limit
    reporting_api.events = [
        {"type": "detected_attack", "time": 1000},
        {"type": "detected_attack", "time": 2000},
        {"type": "detected_attack", "time": 3000},
    ]

    with patch(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
        return_value=4000,
    ):
        response = reporting_api.report(token="token", event=event, timeout_in_sec=5)
        assert response == {"success": False, "error": "max_attacks_reached"}
        assert len(reporting_api.events) == 3  # Should not add the new event


def test_report_within_limit_after_expiry(reporting_api):
    """Test reporting after some events have expired."""
    event1 = {"type": "detected_attack", "time": 1000}
    event2 = {"type": "detected_attack", "time": 2001}

    # Add events to the list
    reporting_api.events = [event1, event2]

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        with patch(
            "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
            return_value=12000,
        ):
            event3 = {"type": "detected_attack", "time": 11000}
            response = reporting_api.report(
                token="token", event=event3, timeout_in_sec=5
            )
            assert response == {"success": True}
            assert (
                len(reporting_api.events) == 2
            )  # One event should have expired, and the new one is added


def test_report_with_non_attack_event(reporting_api):
    """Test reporting with a non-attack event."""
    event = {"type": "other_event", "time": 1000}

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        response = reporting_api.report(token="token", event=event, timeout_in_sec=5)
        assert response == {"success": True}
        assert len(reporting_api.events) == 0  # Non-attack events should not be stored
        mock_report.assert_called_once_with("token", event, 5)


def test_report_multiple_events_within_limit(reporting_api):
    """Test reporting multiple events within the rate limit."""
    events = [
        {"type": "detected_attack", "time": 1000},
        {"type": "detected_attack", "time": 2000},
        {"type": "detected_attack", "time": 3000},
    ]

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        with patch(
            "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
            return_value=4000,
        ):
            for event in events:
                response = reporting_api.report(
                    token="token", event=event, timeout_in_sec=5
                )
                assert response == {"success": True}
            assert len(reporting_api.events) == 3
            assert mock_report.call_count == 3


def test_report_mixed_event_types(reporting_api):
    """Test reporting with a mix of attack and non-attack events."""
    attack_event = {"type": "detected_attack", "time": 1000}
    non_attack_event = {"type": "other_event", "time": 2000}

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        response = reporting_api.report(
            token="token", event=attack_event, timeout_in_sec=5
        )
        assert response == {"success": True}
        assert len(reporting_api.events) == 1

        response = reporting_api.report(
            token="token", event=non_attack_event, timeout_in_sec=5
        )
        assert response == {"success": True}
        assert len(reporting_api.events) == 1  # Non-attack event should not be stored


def test_report_event_expiry(reporting_api):
    """Test that events expire correctly based on the time interval."""
    event1 = {"type": "detected_attack", "time": 1000}
    event2 = {"type": "detected_attack", "time": 2000}
    reporting_api.events = [event1, event2]

    # Simulate time passing
    with patch(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
        return_value=12000,
    ):
        event3 = {"type": "detected_attack", "time": 11000}
        response = reporting_api.report(token="token", event=event3, timeout_in_sec=5)
        assert response == {"error": "timeout", "success": False}
        assert (
            len(reporting_api.events) == 1
        )  # One event should have expired, and the new one is added


def test_report_event_at_boundary(reporting_api):
    """Test reporting an event at the boundary of the interval."""
    event1 = {"type": "detected_attack", "time": 1000}
    reporting_api.events = [event1]

    with patch(
        "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
        return_value=10000,
    ):
        event2 = {"type": "detected_attack", "time": 10000}  # Exactly at the boundary
        response = reporting_api.report(token="token", event=event2, timeout_in_sec=5)
        assert response == {"error": "timeout", "success": False}
        assert (
            len(reporting_api.events) == 2
        )  # Should be added since it's at the boundary


def test_report_invalid_event_type(reporting_api):
    """Test reporting with an invalid event type."""
    event = {"type": "invalid_event", "time": 1000}

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        response = reporting_api.report(token="token", event=event, timeout_in_sec=5)
        assert response == {"success": True}
        assert (
            len(reporting_api.events) == 0
        )  # Invalid event types should not be stored
        mock_report.assert_called_once_with("token", event, 5)


def test_report_no_events(reporting_api):
    """Test reporting when no events have been reported yet."""
    event = {"type": "detected_attack", "time": 1000}

    with patch.object(
        ReportingApiHTTP, "report", return_value={"success": True}
    ) as mock_report:
        with patch(
            "aikido_zen.helpers.get_current_unixtime_ms.get_unixtime_ms",
            return_value=2000,
        ):
            response = reporting_api.report(
                token="token", event=event, timeout_in_sec=5
            )
            assert response == {"success": True}
            assert len(reporting_api.events) == 1  # Should add the first event
            mock_report.assert_called_once_with("token", event, 5)



"""Exports extract_data_from_xml_body helper function"""

import aikido_zen.context as ctx
from aikido_zen.helpers.logging import logger


def extract_data_from_xml_body(user_input, root_element):
    """Extracts all attributes from the xml and adds them to context"""
    try:
        context = ctx.get_current_context()
        if (
            not context
            or not isinstance(context.body, str)
            or user_input != context.body
        ):
            return

        extracted_xml_attrs = context.xml
        for el in root_element:
            for k, v in el.items():
                if not extracted_xml_attrs.get(k):
                    extracted_xml_attrs[k] = set()
                extracted_xml_attrs[k].add(v)
        context.set_as_current_context()
    except Exception as e:
        logger.debug("Exception occurred when extracting XML: %s", e)



import pytest
import requests
from unittest.mock import patch
from aikido_zen.background_process.api.http_api import (
    ReportingApiHTTP,
)  # Replace with the actual module name

# Sample event data for testing
sample_event = {"event_type": "test_event", "data": {"key": "value"}}


def test_report_data_401_code(monkeypatch):
    # Create an instance of ReportingApiHTTP
    api = ReportingApiHTTP("http://mocked-url.com/")

    # Mock the requests.post method to return a successful response
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.json_data = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

    def mock_post(url, json, timeout, headers):
        return MockResponse({"success": False}, 401)

    monkeypatch.setattr(requests, "post", mock_post)

    # Call the report method
    response = api.report("mocked_token", sample_event, 5)

    # Assert the response
    assert response == {"success": False, "error": "invalid_token"}


def test_report_connection_error(monkeypatch):
    # Create an instance of ReportingApiHTTP
    api = ReportingApiHTTP("http://mocked-url.com/")

    # Mock the requests.post method to raise a ConnectionError
    monkeypatch.setattr(
        requests,
        "post",
        lambda *args, **kwargs: (_ for _ in ()).throw(
            requests.exceptions.ConnectionError
        ),
    )

    # Call the report method
    response = api.report("mocked_token", sample_event, 5)

    # Assert the response
    assert response == {"success": False, "error": "timeout"}


def test_report_other_exception(monkeypatch):
    # Create an instance of ReportingApiHTTP
    api = ReportingApiHTTP("http://mocked-url.com/")

    # Mock the requests.post method to raise a generic exception
    def mock_post(url, json, timeout, headers):
        raise Exception("Some error occurred")

    monkeypatch.setattr(requests, "post", mock_post)

    # Call the report method
    response = api.report("mocked_token", sample_event, 5)

    # Assert that the response is None (or however you want to handle it)
    assert response["error"] is "unknown"
    assert not response["success"]



"""Exports get_body_data_type"""

JSON_CONTENT_TYPES = [
    "application/json",
    "application/vnd.api+json",
    "application/csp-report",
    "application/x-json",
]


def get_body_data_type(headers):
    """Gets the type of body data from headers"""
    if not isinstance(headers, dict) or headers is None:
        return

    content_type = headers.get("CONTENT_TYPE")
    if not content_type:
        return

    if isinstance(content_type, list):
        # Choose the first content type if there are multiple
        content_type = content_type[0]

    if any(json_type in content_type for json_type in JSON_CONTENT_TYPES):
        return "json"

    if content_type.startswith("application/x-www-form-urlencoded"):
        return "form-urlencoded"

    if content_type.startswith("multipart/form-data"):
        return "form-data"

    if "xml" in content_type:
        return "xml"

    return None



import pytest
from .contains_unsafe_path_parts import contains_unsafe_path_parts


def test_safe_paths():
    assert not contains_unsafe_path_parts("/home/user/file.txt")
    assert not contains_unsafe_path_parts("C:\\Users\\User\\Documents\\file.txt")
    assert not contains_unsafe_path_parts("C:/Program Files/app.exe")


def test_dangerous_paths():
    assert contains_unsafe_path_parts("/home/user/../file.txt")
    assert contains_unsafe_path_parts("C:\\Users\\User\\..\\Documents\\file.txt")
    assert contains_unsafe_path_parts("..\\..\\file.txt")
    assert contains_unsafe_path_parts("../folder/file.txt")


def test_edge_cases():
    assert not contains_unsafe_path_parts("")
    assert not contains_unsafe_path_parts("..")
    assert not contains_unsafe_path_parts(".")
    assert contains_unsafe_path_parts("folder/../file.txt")
    assert contains_unsafe_path_parts("folder\\..\\file.txt")



import pytest
from .get_body_data_type import get_body_data_type


def test_get_body_data_type():
    assert get_body_data_type({"CONTENT_TYPE": "application/json"}) == "json"
    assert get_body_data_type({"CONTENT_TYPE": "application/vnd.api+json"}) == "json"
    assert get_body_data_type({"CONTENT_TYPE": "application/csp-report"}) == "json"
    assert get_body_data_type({"CONTENT_TYPE": "application/x-json"}) == "json"
    assert (
        get_body_data_type({"CONTENT_TYPE": "application/x-www-form-urlencoded"})
        == "form-urlencoded"
    )
    assert get_body_data_type({"CONTENT_TYPE": "multipart/form-data"}) == "form-data"
    assert get_body_data_type({"CONTENT_TYPE": "text/xml"}) == "xml"
    assert get_body_data_type({"CONTENT_TYPE": "text/html"}) is None
    assert (
        get_body_data_type({"CONTENT_TYPE": ["application/json", "text/html"]})
        == "json"
    )
    assert get_body_data_type({"x-test": "abc"}) is None
    assert get_body_data_type(None) is None  # Testing invalid input
    assert get_body_data_type({}) is None



"""Commands __init__.py file"""

from aikido_zen.helpers.logging import logger
from .attack import process_attack
from .read_property import process_read_property
from .should_ratelimit import process_should_ratelimit
from .ping import process_ping
from .sync_data import process_sync_data

commands_map = {
    # This maps to a tuple : (function, returns_data?)
    # Commands that don't return data :
    "ATTACK": (process_attack, False),
    # Commands that return data :
    "SYNC_DATA": (process_sync_data, True),
    "READ_PROPERTY": (process_read_property, True),
    "SHOULD_RATELIMIT": (process_should_ratelimit, True),
    "PING": (process_ping, True),
}


def process_incoming_command(connection_manager, obj, conn, queue):
    """Processes an incoming command"""
    action = obj[0]
    data = obj[1]
    if action in commands_map:
        func, returns_data = commands_map[action]
        if returns_data:
            return conn.send(func(connection_manager, data, queue))
        func(connection_manager, data, queue)
    else:
        logger.debug("Command : `%s` not found, aborting", action)



"""Exports process_renew_config"""

from aikido_zen.api_discovery.update_route_info import update_route_info
from aikido_zen.helpers.logging import logger


def process_sync_data(connection_manager, data, conn, queue=None):
    """
    Synchronizes data between the thread-local cache (with a TTL of usually 1 minute) and the
    background thread. Which data gets synced?
    Thread -> BG Process : Routes, Hostnames, Users, Stats & middleware installed
    BG Process -> Thread : Routes and config
    """

    # Sync routes
    routes = connection_manager.routes
    for route in data.get("current_routes", {}).values():
        route_metadata = {"method": route["method"], "route": route["path"]}
        if not routes.get(route_metadata):
            routes.initialize_route(route_metadata)
        existing_route = routes.get(route_metadata)

        # Update hit count :
        hits_delta_since_sync = int(route.get("hits_delta_since_sync", 0))
        existing_route["hits"] += hits_delta_since_sync

        # Update API Spec :
        update_route_info(route["apispec"], existing_route)

    # Save middleware installed :
    if data.get("middleware_installed", False):
        connection_manager.middleware_installed = True

    # Sync hostnames
    for hostnames_entry in data.get("hostnames", list()):
        connection_manager.hostnames.add(
            hostnames_entry["hostname"],
            hostnames_entry["port"],
            hostnames_entry["hits"],
        )

    # Sync users
    for user_entry in data.get("users", list()):
        connection_manager.users.add_user_from_entry(user_entry)

    # Sync stats
    connection_manager.statistics.import_from_record(data.get("stats", {}))

    # Sync ai stats
    connection_manager.ai_stats.import_list(data.get("ai_stats", []))

    if connection_manager.conf.last_updated_at > 0:
        # Only report data if the config has been fetched.
        return {
            "routes": dict(connection_manager.routes.routes),
            "config": connection_manager.conf,
        }
    return {}



import pytest
from unittest.mock import Mock, patch
from .should_ratelimit import process_should_ratelimit


@pytest.mark.parametrize(
    "should_ratelimit, expected_call",
    [
        (True, True),
        (False, False),
    ],
)
def test_process_should_ratelimit(should_ratelimit, expected_call):
    """Test rate limiting behavior based on different return values."""
    # Arrange
    connection_manager = Mock()
    data = {
        "route_metadata": {
            "route": "/test",
            "url": "http://example.com/test",
            "method": "GET",
        },
        "remote_address": "192.168.1.1",
        "user": {"id": 1, "name": "Test User"},
    }

    with patch(
        "aikido_zen.ratelimiting.should_ratelimit_request",
        return_value=should_ratelimit,
    ) as mock_should_ratelimit:
        # Act
        process_should_ratelimit(connection_manager, data)

    # Assert
    mock_should_ratelimit.assert_called_once_with(
        route_metadata=data["route_metadata"],
        remote_address=data["remote_address"],
        user=data["user"],
        connection_manager=connection_manager,
    )


def test_process_should_ratelimit_no_connection_manager():
    """Test that process_should_ratelimit does nothing when connection_manager is not present."""
    # Arrange
    data = {
        "route_metadata": {
            "route": "/test",
            "url": "http://example.com/test",
            "method": "GET",
        },
        "remote_address": "192.168.1.1",
        "user": {"id": 1, "name": "Test User"},
    }

    # Act
    result = process_should_ratelimit(None, data)  # No connection_manager

    # Assert
    assert result is False  # Ensure the function returns False


def test_process_should_ratelimit_multiple_calls():
    """Test multiple calls to process_should_ratelimit."""
    # Arrange
    connection_manager = Mock()
    data = {
        "route_metadata": {
            "route": "/test",
            "url": "http://example.com/test",
            "method": "GET",
        },
        "remote_address": "192.168.1.1",
        "user": {"id": 1, "name": "Test User"},
    }

    with patch(
        "aikido_zen.ratelimiting.should_ratelimit_request",
        side_effect=[True, False],
    ) as mock_should_ratelimit:
        # Act
        process_should_ratelimit(connection_manager, data)  # First call
        process_should_ratelimit(connection_manager, data)  # Second call

    # Assert
    mock_should_ratelimit.assert_any_call(
        route_metadata=data["route_metadata"],
        remote_address=data["remote_address"],
        user=data["user"],
        connection_manager=connection_manager,
    )


def test_process_should_ratelimit_with_different_connection_manager():
    """Test with different connection_manager configurations."""
    # Arrange
    connection_manager1 = Mock(name="CloudConnectionManager1")
    connection_manager2 = Mock(name="CloudConnectionManager2")
    data = {
        "route_metadata": {
            "route": "/test",
            "url": "http://example.com/test",
            "method": "GET",
        },
        "remote_address": "192.168.1.1",
        "user": {"id": 1, "name": "Test User"},
    }

    with patch(
        "aikido_zen.ratelimiting.should_ratelimit_request", return_value=True
    ) as mock_should_ratelimit:
        # Act
        process_should_ratelimit(connection_manager1, data)
        process_should_ratelimit(connection_manager2, data)

    # Assert
    assert mock_should_ratelimit.call_count == 2
    mock_should_ratelimit.assert_any_call(
        route_metadata=data["route_metadata"],
        remote_address=data["remote_address"],
        user=data["user"],
        connection_manager=connection_manager1,
    )
    mock_should_ratelimit.assert_any_call(
        route_metadata=data["route_metadata"],
        remote_address=data["remote_address"],
        user=data["user"],
        connection_manager=connection_manager2,
    )



import pytest
import socket
from aikido_zen.helpers.get_machine_ip import get_ip


def test_get_ip_success(monkeypatch):
    # Mock the socket.gethostname to return a specific hostname
    monkeypatch.setattr(socket, "gethostname", lambda: "mocked_hostname")
    # Mock the socket.gethostbyname to return a specific IP address
    monkeypatch.setattr(socket, "gethostbyname", lambda hostname: "192.168.1.1")

    assert get_ip() == "192.168.1.1"


def test_get_ip_failure(monkeypatch):
    # Mock the socket.gethostname to return a specific hostname
    monkeypatch.setattr(socket, "gethostname", lambda: "mocked_hostname")
    # Mock the socket.gethostbyname to raise an exception
    monkeypatch.setattr(
        socket,
        "gethostbyname",
        lambda hostname: (_ for _ in ()).throw(Exception("Mocked exception")),
    )

    assert get_ip() == "x.x.x.x"



"""Exports `ip_allowed_to_access_route`"""

from aikido_zen.helpers.is_localhost_ip import is_localhost_ip
from aikido_zen.helpers.iplist import IPList


def ip_allowed_to_access_route(remote_address, route_metadata, endpoints):
    """
    Checks if the ip address can access the route, given the service conf
    """
    ip = remote_address
    if ip and is_localhost_ip(ip):
        return True

    if not endpoints:
        return True

    for endpoint in endpoints:
        if not "allowedIPAddresses" in endpoint:
            #  This feature is not supported by the current aikido server version
            continue
        if not isinstance(endpoint["allowedIPAddresses"], IPList):
            #  We will continue to check all the other matches
            continue
        if not ip:
            # We only check it here because if allowedIPAddresses isn't set
            # We don't want to change any default behaviour
            return False
        if not endpoint["allowedIPAddresses"].matches(ip):
            # The IP is not in the allowlist, so block
            return False
    return True



"""
Helper function file, see function definition for more info
"""


def get_current_and_next_segments(array):
    """Get the current and next segments of an array"""
    return [(array[i], array[i + 1]) for i in range(len(array) - 1)]



import pytest
from .find_hostname_in_userinput import find_hostname_in_userinput


def test_returns_false_if_user_input_and_hostname_are_empty():
    assert find_hostname_in_userinput("", "") is False


def test_returns_false_if_user_input_is_empty():
    assert find_hostname_in_userinput("", "example.com") is False


def test_returns_false_if_hostname_is_empty():
    assert find_hostname_in_userinput("http://example.com", "") is False


def test_it_parses_hostname_from_user_input():
    assert find_hostname_in_userinput("http://localhost", "localhost") is True


def test_it_parses_special_ip():
    assert find_hostname_in_userinput("http://localhost", "localhost") is True


def test_it_parses_hostname_from_user_input_with_path_behind_it():
    assert find_hostname_in_userinput("http://localhost/path", "localhost") is True


def test_it_does_not_parse_hostname_from_user_input_with_misspelled_protocol():
    assert find_hostname_in_userinput("http:/localhost", "localhost") is False


def test_it_does_not_parse_hostname_from_user_input_without_protocol_separator():
    assert find_hostname_in_userinput("http:localhost", "localhost") is False


def test_it_does_not_parse_hostname_from_user_input_with_misspelled_protocol_and_path_behind_it():
    assert find_hostname_in_userinput("http:/localhost/path/path", "localhost") is False


def test_it_parses_hostname_from_user_input_without_protocol_and_path_behind_it():
    assert find_hostname_in_userinput("localhost/path/path", "localhost") is True


def test_it_flags_ftp_as_protocol():
    assert find_hostname_in_userinput("ftp://localhost", "localhost") is True


def test_it_parses_hostname_from_user_input_without_protocol():
    assert find_hostname_in_userinput("localhost", "localhost") is True


def test_it_ignores_invalid_urls():
    assert find_hostname_in_userinput("http://", "localhost") is False


def test_user_input_is_smaller_than_hostname():
    assert find_hostname_in_userinput("localhost", "localhost localhost") is False


def test_it_finds_ip_address_inside_url():
    assert (
        find_hostname_in_userinput(
            "http://169.254.169.254/latest/meta-data/", "169.254.169.254"
        )
        is True
    )


def test_it_finds_ip_address_with_strange_notation_inside_url():
    assert find_hostname_in_userinput("http://2130706433", "2130706433") is True
    assert find_hostname_in_userinput("http://127.1", "127.1") is True
    assert find_hostname_in_userinput("http://127.0.1", "127.0.1") is True


def test_it_works_with_invalid_ports():
    assert (
        find_hostname_in_userinput(
            "http://localhost:1337\\u0000asd.php", "localhost", 1337
        )
        is True
    )


def test_it_works_with_ports():
    assert find_hostname_in_userinput("http://localhost", "localhost", 8080) is False
    assert (
        find_hostname_in_userinput("http://localhost:8080", "localhost", 8080) is True
    )
    assert find_hostname_in_userinput("http://localhost:8080", "localhost") is True
    assert (
        find_hostname_in_userinput("http://localhost:8080", "localhost", 4321) is False
    )



"""Re-exports middleware"""

from .asgi import AikidoASGIMiddleware as AikidoQuartMiddleware
from .asgi import AikidoASGIMiddleware as AikidoStarletteMiddleware
from .asgi import AikidoASGIMiddleware as AikidoFastAPIMiddleware
from .flask import AikidoFlaskMiddleware
from .django import AikidoDjangoMiddleware
from .should_block_request import should_block_request
